try:
    import h5py
except:
    pass

import logging
import numpy as np
from fastlmm.pyplink.snpset import *
from fastlmm.pyplink.altset_list import *

#!!document the format


class Hdf5(object):
    def __init__(self, filename, order="F", blocksize=5000):
        self._ran_once = False
        self.h5 = None

        ##!! copy relevent comments from Bed reader
        self.filename = filename
        self.order = order
        self.blocksize = blocksize

    def copyinputs(self, copier):
        copier.input(self.filename)

    @property
    def snp_to_index(self):
        self.run_once()
        return self._snp_to_index

    def run_once(self):
        if self._ran_once:
            return
        self._ran_once = True
        try:
            self.h5 = h5py.File(self.filename, "r")
        except IOError as e:
            raise IOError(
                "Missing or unopenable file '{0}' -- Native error message: {1}".format(
                    self.filename, e
                )
            )

        self._original_iids = np.array(np.array(self.h5["iid"]), dtype=str)
        self.rs = np.array(np.array(self.h5["rs"]), dtype="str")
        self.pos = np.array(self.h5["pos"])

        ## similar code in bed
        self._snp_to_index = {}
        logging.info("indexing snps")
        for i, snp in enumerate(self.rs):
            if snp in self._snp_to_index:
                raise Exception(
                    "Expect snp to appear in bim file only once. ({0})".format(snp)
                )
            self._snp_to_index[snp] = i

        self.snpsInFile = self.h5["snps"]

        if "SNP-major" not in self.snpsInFile.attrs:
            raise Exception(
                "In Hdf5 the 'snps' matrix must have a Boolean 'SNP-major' attribute"
            )
        self.is_snp_major = self.snpsInFile.attrs["SNP-major"]

        S_original = len(self.rs)
        N_original = len(self.original_iids)
        if self.is_snp_major:
            if not self.snpsInFile.shape == (S_original, N_original):
                raise Exception(
                    "In Hdf5, snps matrix dimensions don't match those of 'rs' and 'iid'"
                )
        else:
            if not self.snpsInFile.shape == (N_original, S_original):
                raise Exception(
                    "In Hdf5, snps matrix dimensions don't match those of 'rs' and 'iid'"
                )

    @property
    def snp_count(self):
        self.run_once()
        return len(self.rs)

    @property
    def original_iids(self):
        self.run_once()
        return self._original_iids

    # same code is in Bed. Could this be moved to an abstract class?
    def read(
        self, snp_set=AllSnps(), order="F", dtype=np.float64, force_python_only=False
    ):
        self.run_once()
        snpset_withbed = snp_set.addbed(self)
        return self.read_with_specification(
            snpset_withbed,
            order=order,
            dtype=dtype,
            force_python_only=force_python_only,
        )

    @staticmethod
    # should move into utils
    def is_sorted_without_repeats(list):
        if len(list) < 2:
            return True
        for i in range(1, len(list)):
            if not list[i - 1] < list[i]:
                return False
        return True

    def __del__(self):
        if (
            self.h5 is not None
        ):  # we need to test this because Python doesn't guarantee that __init__ was fully run
            self.h5.close()

    def read_direct(self, snps, selection=np.s_[:, :]):
        if self.is_snp_major:
            selection = tuple(reversed(selection))

        if snps.flags["F_CONTIGUOUS"]:
            self.snpsInFile.read_direct(snps.T, selection)
        else:
            self.snpsInFile.read_direct(snps, selection)

    #!! much code the same as for Bed
    def create_block(self, blocksize, dtype, order):
        N_original = len(self.original_iids)  # similar code else where -- make a method
        matches_order = self.is_snp_major == (
            order == "F"
        )  # similar code else where -- make a method
        opposite_order = (
            "C" if order == "F" else "F"
        )  # similar code else where -- make a method
        if matches_order:
            return np.empty([N_original, blocksize], dtype=dtype, order=order)
        else:
            return np.empty([N_original, blocksize], dtype=dtype, order=opposite_order)

    def read_with_specification(
        self,
        snpset_with_snpreader,
        order="F",
        dtype=np.float64,
        force_python_only=False,
    ):
        self.run_once()

        order = order.upper()
        opposite_order = "C" if order == "F" else "F"

        snp_index_list = np.array(
            list(snpset_with_snpreader)
        )  # Is there a way to create an array from an iterator without putting it through a list first?
        S = len(snp_index_list)
        S_original = self.snp_count
        N_original = len(self.original_iids)

        # Check if snps and iids indexes are in order and in range
        snps_are_sorted = Hdf5.is_sorted_without_repeats(snp_index_list)
        if hasattr(self, "_ind_used"):
            iid_index_list = self._ind_used
            iid_is_sorted = Hdf5.is_sorted_without_repeats(iid_index_list)
        else:
            iid_index_list = np.arange(N_original)
            iid_is_sorted = True

        N = len(iid_index_list)

        SNPs = np.empty([N, S], dtype=dtype, order=order)

        matches_order = self.is_snp_major == (order == "F")
        is_simple = (
            not force_python_only
            and iid_is_sorted
            and snps_are_sorted
            and matches_order
        )  # If 'is_simple' may be able to use a faster reader

        # case 1 - all snps & all ids requested
        if is_simple and S == S_original and N == N_original:
            self.read_direct(SNPs)

        # case 2 - some snps and all ids
        elif is_simple and N == N_original:
            self.read_direct(SNPs, np.s_[:, snp_index_list])

        # case 3 all snps and some ids
        elif is_simple and S == S_original:
            self.read_direct(SNPs, np.s_[iid_index_list, :])

        # case 4 some snps and some ids -- use blocks
        else:
            blocksize = min(self.blocksize, S)
            block = self.create_block(blocksize, dtype, order)

            if not snps_are_sorted:
                snp_index_index_list = np.argsort(snp_index_list)
                snp_index_list_sorted = snp_index_list[snp_index_index_list]
            else:
                snp_index_index_list = np.arange(S)
                snp_index_list_sorted = snp_index_list

            for start in range(0, S, blocksize):
                # print start
                end = min(start + blocksize, S)
                if (
                    end - start < blocksize
                ):  # On the last loop, the buffer might be too big, so make it smaller
                    block = self.create_block(end - start, dtype, order)
                snp_index_list_forblock = snp_index_list_sorted[start:end]
                snp_index_index_list_forblock = snp_index_index_list[start:end]
                self.read_direct(block, np.s_[:, snp_index_list_forblock])
                SNPs[:, snp_index_index_list_forblock] = block[iid_index_list, :]

        rs = self.rs[snp_index_list]
        pos = self.pos[snp_index_list, :]
        iids = np.array(
            self.original_iids[iid_index_list], dtype="str"
        )  # Need to make another copy of to stop it from being converted to a list of 1-d string arrays

        has_right_order = (order == "C" and SNPs.flags["C_CONTIGUOUS"]) or (
            order == "F" and SNPs.flags["F_CONTIGUOUS"]
        )
        # if SNPs.shape == (1, 1):
        assert SNPs.shape == (N, S) and SNPs.dtype == dtype and has_right_order

        ret = {"rs": rs, "pos": pos, "snps": SNPs, "iid": iids}
        return ret

    @property
    def ind_used(self):
        # doesn't need to self.run_once() because only uses original inputs
        return self._ind_used

    @ind_used.setter
    def ind_used(self, value):
        """
        Tell the Bed reader to return data for only a subset (perhaps proper) of the individuals in a particular order
        e.g. 2,10,0 says to return data for three users: the user at index position 2, the user at index position 10, and the user at index position 0.
        """
        # doesn't need to self.run_once() because only uses original inputs
        self._ind_used = value

    @staticmethod
    def write(snpMatrix, hdf5file, dtype="f8", snp_major=True, compression=None):
        if not isinstance(dtype, str) or len(dtype) != 2 or dtype[0] != "f":
            raise Exception(
                "Expect dtype to start with 'f', e.g. 'f4' for single, 'f8' for double"
            )
        data = (snpMatrix["snps"].T) if snp_major else snpMatrix["snps"]
        with h5py.File(hdf5file, "w") as h5:
            h5.create_dataset(
                "snps", data=data, dtype=dtype, compression=compression, shuffle=True
            )
            h5["snps"].attrs["SNP-major"] = snp_major
            h5.create_dataset("iid", data=snpMatrix["iid"])
            h5.create_dataset("pos", data=snpMatrix["pos"])
            h5.create_dataset("rs", data=snpMatrix["rs"])



import os
import logging
import unittest
from pysnptools.util.filecache import Hashdown
import doctest


fastlmm_hashdown = Hashdown.load_hashdown(
    os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "fastlmm.hashdown.json"
    ),
    directory=os.environ.get("FASTKNN_CACHE_HOME", None),
)

def example_file(pattern, endswith=None):
    """
    Returns the local location of a FaST-LMM example file, downloading it
    if needed.

    :param pattern: The name of the example file of interest. A 
        `file name pattern <https://docs.python.org/3.7/library/fnmatch.html>`__
        may be given. All matching files will be downloaded (if needed) and
        the name of one will be returned.
    :type file_name: string

    :param endswith: The pattern of the file name to return. By default, when
        no `endswith` is given, the name of the first matching file will be
        returned.
    :type file_name: string

    :rtype: string

    By default, the local location will be under the system temp directory
    (typically controlled with the TEMP environment variable).
    Alternatively, the directory can be set with the FASTLMM_CACHE_HOME
    environment variable.

    This function knows the MD5 hash of all FaST-LMM example files and uses
    that content-based hash to decide if a file needs to be downloaded.

    >>> from fastlmm.util import example_file # Download and return local file name # Download and return local file name
    >>> # Download the phenotype file if necessary. Return its local location.
    >>> pheno_fn = example_file("fastlmm/feature_selection/examples/toydata.phe")
    >>> print('The local file name is ', pheno_fn)
    The local file name is ...fastlmm/feature_selection/examples/toydata.phe
    >>>
    >>> # Download the bed,bim,&fam files if necessary. Return the location of bed file.
    >>> bedfile = example_file("fastlmm/feature_selection/examples/toydata.5chrom.*","*.bed")
    >>> print('The local file name is ', bedfile)
    The local file name is ...fastlmm/feature_selection/examples/toydata.5chrom.bed

    """
    return fastlmm_hashdown._example_file(pattern, endswith=endswith)


class TestExampleFile(unittest.TestCase):
    def test_doc_test(self):
        import fastlmm.util._example_file

        result = doctest.testmod(
            fastlmm.util._example_file, optionflags=doctest.ELLIPSIS
        )
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestExampleFile))

    return test_suite


if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import re
import numpy as NP
import scipy.io as SIO
import time
import os
import os.path
import sys
from fastlmm.association.FastLmmSet import FastLmmSet
from fastlmm.association.FastLmmSet import Local
import unittest
import subprocess
import fastlmm.inference.tests.test
import fastlmm.feature_selection.test
import pysnptools.test
import shutil
import logging
import fastlmm.util.util as ut
from pysnptools.util.mapreduce1.distributabletest import DistributableTest
from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc, LocalInParts
from fastlmm.pyplink.snpreader.Hdf5 import Hdf5  # Need by some *.py's that we eval

tolerance = 1e-4

# def compare_files(afilename, bfilename):
#    if not os.path.isfile(afilename) or not os.path.isfile(bfilename):
#        return False
#    if os.name == 'posix':
#        call = 'wine CompareFiles/CompareFiles.exe'
#    else:
#        call = 'CompareFiles/CompareFiles.exe'
#    out = subprocess.check_output(call+' -tol '+str(tolerance)+' '+afilename+' '
#                                  +bfilename, shell=True)
#    return out.find('Files are comparable.') != -1


class WidgetTestCase(unittest.TestCase):
    def __init__(self, infile):
        unittest.TestCase.__init__(self)

        self._infile = infile
        # if not self._existExpectedOutput():
        #    self._generateExpectedOutput()

    def _tmpOutfile(self):
        outfile = os.path.splitext(self._infile)[0]
        return "tmp/" + outfile + ".txt"

    def _referenceOutfile(self):
        #!!similar code elsewhere
        import platform

        os_string = platform.platform()
        outfile = os.path.splitext(self._infile)[0]

        windows_fn = "expected-Windows/" + outfile + ".txt"
        assert os.path.exists(windows_fn), "Can't find file '{0}'".format(windows_fn)
        debian_fn = "expected-debian/" + outfile + ".txt"
        if not os.path.exists(
            debian_fn
        ):  # If reference file is not in debian folder, look in windows folder
            debian_fn = windows_fn

        if "debian" in os_string or "Linux" in os_string:
            if "Linux" in os_string:
                logging.warning(
                    "comparing to Debian output even though found: %s" % os_string
                )
            return debian_fn
        else:
            if "Windows" not in os_string:
                logging.warning(
                    "comparing to Windows output even though found: %s" % os_string
                )
            return windows_fn

    def runTest(self):
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        tmpOutfile = self._tmpOutfile()
        referenceOutfile = self._referenceOutfile()
        logging.info(f"{'inputs/'+self._infile}")
        with open("inputs/" + self._infile) as f:
            filecontent = f.read()

        runner = Local()
        try:
            distributable = eval(filecontent)
        except Exception as e:
            raise Exception(
                "Can't eval '{0}' because of '{1}'".format(self._infile, e.message)
            )
        runner.run(distributable)

        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out,
            "msg='{0}', ref='{1}', tmp='{2}'".format(msg, referenceOutfile, tmpOutfile),
        )

    # def _generateExpectedOutput(self):
    #    tmpOutfile = self._tmpOutfile()
    #    referenceOutfile = self._referenceOutfile()
    #    with open('inputs/'+self._infile) as f:
    #        filecontent = f.read()

    #    runner = Local()
    #    exec(filecontent)
    #    runner.run(distributable)

    #    shutil.copyfile(self._tmpOutfile(), self._referenceOutfile())

    def _existExpectedOutput(self):
        return os.path.isfile(self._referenceOutfile())

    def __str__(self):
        return self._infile


##for debuggingk
# def getDebugTestSuite():
#    suite = unittest.TestSuite()
#    suite.addTest(WidgetTestCase('sc_davies_two_kernel_linear_qqfit.N300.needsReorder.py'))
#    return suite


def getTestSuiteX():
    suite = unittest.TestSuite()
    for f in os.listdir("inputs"):
        if re.match(r".*\.py$", f) is None:
            continue
        # from tests.test import WidgetTestCase
        suite.addTest(WidgetTestCase(f))
    return suite


def removeTmpFiles():
    if os.path.exists("tmp"):
        shutil.rmtree("tmp")
        # for f in os.listdir( 'tmp' ):
        #    os.remove(os.path.join('tmp', f), )


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARN)
    removeTmpFiles()

    import fastlmm.association.tests.testepistasis
    import fastlmm.association.tests.test_single_snp
    import fastlmm.association.tests.test_single_snp_linreg
    import fastlmm.association.tests.test_single_snp_select
    import fastlmm.association.tests.test_single_snp_all_plus_select
    import fastlmm.association.tests.test_snp_set
    import fastlmm.association.tests.test_gwas
    import fastlmm.association.tests.test_heritability_spatial_correction
    import fastlmm.association.tests.test_single_snp_scale
    import fastlmm.inference.tests.test_fastlmm_predictor
    import fastlmm.inference.tests.test_linear_regression
    import fastlmm.inference.tests.test
    import fastlmm.util.test
    import fastlmm.pyplink.test

    suites = unittest.TestSuite(
        [
            # cmk
            # for debugging getDebugTestSuite(),
            getTestSuiteX(),
            fastlmm.inference.tests.test_linear_regression.getTestSuite(),
            fastlmm.association.tests.test_single_snp_scale.getTestSuite(),
            fastlmm.pyplink.test.getTestSuite(),
            fastlmm.feature_selection.test.getTestSuite(),
            # this test is deprecated # fastlmm.association.tests.test_single_snp_select.getTestSuite(),
            fastlmm.inference.tests.test_fastlmm_predictor.getTestSuite(),
            fastlmm.association.tests.test_gwas.getTestSuite(),
            fastlmm.association.tests.test_snp_set.getTestSuite(),
            fastlmm.inference.tests.test.getTestSuite(),
            fastlmm.association.tests.testepistasis.getTestSuite(),
            fastlmm.association.tests.test_heritability_spatial_correction.getTestSuite(),
            fastlmm.util.test.getTestSuite(),
            fastlmm.association.tests.test_single_snp_all_plus_select.getTestSuite(),
            fastlmm.inference.tests.test.getTestSuite(),
            fastlmm.association.tests.test_single_snp.getTestSuite(),
            fastlmm.association.tests.test_single_snp_linreg.getTestSuite(),
        ]
    )

    if True:  # Standard test run
        r = unittest.TextTestRunner(failfast=True)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else:  # Cluster test run
        # Because both pysnsptools and fastlmm contain a tests folder, to run on cluster must have fastlmm listed first in the PYTHONPATH

        runner = Local()
        runner = LocalMultiProc(taskcount=12, mkl_num_threads=5, just_one_process=False)
        # runner = LocalInParts(1,2,mkl_num_threads=1) # For debugging the cluster runs
        distributable_test = DistributableTest(suites, "temp_test")
        runner.run(distributable_test)

    debian_count = len(os.listdir("expected-debian"))
    if debian_count > 0:
        logging.warning(
            "The tests contain {0} expected-results files that differ between Windows and Debian".format(
                debian_count
            )
        )

    logging.info("done with testing")



import warnings


class OutputWriter(object):
    """Centralize output without messing with sys.stdout."""

    _streams = []

    def addOutputStream(self, out):
        for o in self._streams:
            if o is out:
                return
        self._streams.append(out)

    def write(self, msg):
        warnings.warn(
            "Pr and OutputWriter are deprecated. Use the standard logging.info() instead",
            DeprecationWarning,
        )
        for s in self._streams:
            s.write(msg)

    def prin(self, msg):
        warnings.warn(
            "Pr and OutputWriter are deprecated. Use the standard logging.info() instead",
            DeprecationWarning,
        )
        self.write(msg + "\n")

    def close(self):
        for s in self._streams:
            s.close()

    def fileno(self):
        for s in self._streams:
            if "fileno" in dir(s):
                return s.fileno()
        return 0

    def flush(self):
        for s in self._streams:
            if "flush" in dir(s):
                s.flush()


# This is the only OutputWriter instance that you should use.
Pr = OutputWriter()



import numpy as np
import subprocess, sys, os.path


class SnpAndSetName(object):  # implements ISnpSet
    """
    See the Bed class's 'read' method of examples of its use.
    See __init__.py for specification of interface it implements.
    """

    def __init__(self, setname, snplist):
        self.name = setname
        self.snplist = snplist

    def addbed(self, bed):
        return SnpAndSetNamePlusBed(self.name, self.snplist, bed)


class SnpAndSetNamePlusBed(object):  # implements ISnpSetPlusBed
    """
    A single set of snps.
    """

    def __init__(self, setname, snplist, bed):
        self.name = setname
        self.snplist = [
            val for val in snplist if val in bed.snp_to_index
        ]  # use the intersection
        self.bed = bed

    def __str__(self):
        return self.name

    def __len__(self):
        return len(self.snplist)

    def __iter__(self):
        for snp in self.snplist:
            index = self.bed.snp_to_index[snp]
            yield index

    def read(self):  #!!why don't all the interface implementers have this method?
        return self.bed.read_with_specification(self)

    @property
    def pos(self):
        """
        Returns:
            pos:    position of the SNPs in the specification
        """
        return self.bed.pos[self.to_index]

    @property
    def to_index(self):
        iter = self.__iter__()
        return [i for i in iter]



from .Beta import *
from .Unit import *

# import warnings
# warnings.warn("This __init__.py is deprecated. Pysnptools includes newer version", DeprecationWarning)


def factory(s):
    s = s.capitalize()
    if s == "Unit" or s == "Unit()":
        return Unit()

    if s == "Beta":
        return Beta()

    if s.startswith("Beta("):
        standardizer = eval(s)
        return standardizer


def standardize_with_lambda(snps, lambdax, blocksize=None):
    if blocksize == None:
        return lambdax(snps)

    idx_start = 0
    idx_stop = blocksize

    while idx_start < snps.shape[1]:
        # print idx_start
        lambdax(snps[:, idx_start:idx_stop])

        idx_start = idx_stop
        idx_stop += blocksize
        if idx_stop > snps.shape[1]:
            idx_stop = snps.shape[1]

    return snps


def standardize_unit_python(snps, returnStats=False):
    """
    standardize snps to zero-mean and unit variance
    """

    N = snps.shape[0]
    S = snps.shape[1]

    imissX = np.isnan(snps)
    snp_sum = np.nansum(snps, axis=0)
    n_obs_sum = (~imissX).sum(0)

    snp_mean = (snp_sum * 1.0) / n_obs_sum
    snps -= snp_mean
    snp_std = np.sqrt(np.nansum(snps**2, axis=0) / n_obs_sum)

    # avoid div by 0 when standardizing
    if snp_std.__contains__(0.0):
        logging.warning(
            "A least one snps has only one value, that is, its standard deviation is zero"
        )
    snp_std[snp_std == 0.0] = 1.0
    snps /= snp_std
    snps[imissX] = 0

    if returnStats:
        return snps, snp_mean, snp_std

    return snps


def standardize_beta_python(snps, betaA, betaB):
    """
    standardize snps with Beta prior
    """

    N = snps.shape[0]
    S = snps.shape[1]

    imissX = np.isnan(snps)
    snp_sum = np.nansum(snps, axis=0)
    n_obs_sum = (~imissX).sum(0)

    snp_mean = (snp_sum * 1.0) / n_obs_sum
    snps -= snp_mean
    snp_std = np.sqrt(np.nansum(snps**2, axis=0) / n_obs_sum)
    if snp_std.__contains__(0.0):
        logging.warning(
            "A least one snps has only one value, that is, its standard deviation is zero"
        )

    maf = snp_mean / 2.0
    maf[maf > 0.5] = 1.0 - maf[maf > 0.5]

    # avoid div by 0 when standardizing
    import scipy.stats as st

    maf_beta = st.beta.pdf(maf, betaA, betaB)
    snps *= maf_beta
    snps[imissX] = 0.0

    return snps



import numpy as np
import subprocess, sys, os.path
from itertools import *
import logging


class SnpAndSetNameCollection(object):  # implements ISnpSetList
    """
    Specifies a list of snp sets via a file that has columns 'snp' and 'group'.
    See the Bed class's 'read' method of examples of its use.
    See __init__.py for specification of interface it implements.
    """

    def __init__(self, filename):
        self.filename = filename
        logging.info("Reading {0}".format(filename))
        import pandas as pd

        snp_and_setname_sequence = pd.read_csv(
            filename, delimiter="\t", index_col=False
        )

        from collections import defaultdict

        setname_to_sid = defaultdict(list)
        for snp, gene in snp_and_setname_sequence.itertuples(index=False):
            setname_to_sid[gene].append(snp)
        self.bigToSmall = sorted(
            setname_to_sid.items(), key=lambda gene_sid: -len(gene_sid[1])
        )

    def addbed(self, bed):
        return _SnpAndSetNameCollectionPlusBed(self, bed)

    def copyinputs(self, copier):
        copier.input(self.filename)

    # would be nicer if these used generic pretty printer
    def __repr__(self):
        return "SnpAndSetNameCollection(filename={0})".format(self.filename)

    def __iter__(self):
        for gene, sid in self.bigToSmall:
            yield gene, sid


class _SnpAndSetNameCollectionPlusBed(object):  # implements ISnpSetListPlusBed
    """
    The SnpAndSetNameCollection with the addition of BED information.
    """

    def __init__(self, spec, bed):
        self.spec = spec
        self.bed = bed

    def __len__(self):
        return len(self.spec.bigToSmall)

    def __iter__(self):
        for gene, sid in self.spec.bigToSmall:
            if len(set(sid)) != len(sid):
                raise Exception(
                    "Some snps in gene {0} are listed more than once".format(gene)
                )
            sid_index = self.bed.sid_to_index(sid)
            yield gene, self.bed[:, sid_index]



import numpy as np
import scipy.linalg as LA
import numpy.linalg as nla
import os
import sys
import glob

sys.path.append("./../../pyplink")
from fastlmm.pyplink.plink import *
from pysnptools.util.pheno import *
from fastlmm.util.mingrid import *

# import pdb
import scipy.stats as ST
import fastlmm.util.stats as ss
import fastlmm.util.util as util
import fastlmm.association as association


class scoretest(association.varcomp_test):
    """
    This is the super class that just performs a score test for the 1K linear case, gives P-values etc.
    All other models are inherited
    """

    __slots__ = ["squaredform", "expectationsqform", "varsqform", "GPG", "GPY"]

    def __init__(self, Y, X=None, appendbias=False):
        association.varcomp_test.__init__(self, Y=Y, X=X, appendbias=appendbias)
        pass

    def _score(self, G1):
        """
        This calls the score computation for a single kernel
        Christoph guess: varsqform is the variance of sigma_g, and is the inverse of the Fisher information wrt sigma_g
        Christoph: should compute variance of h2 (and test h2>0), which eliminates one nusiance parameter and yields a better test
        """
        (
            self.squaredform,
            self.expectationsqform,
            self.varsqform,
            self.GPG,
            self.GPY,
        ) = scoreNoK(
            Y=self.Y, X=self.X, Xdagger=None, G=G1, sigma2=None, Bartletcorrection=True
        )
        if self.GPG.shape[0] == 0:
            raise Exception("GPG is empty")
        return self.squaredform, self.expectationsqform, self.varsqform, self.GPG

    # def _pv(self, type): # this used to default to ="davies"
    #    evalstring = 'self.pv_%s(self.squaredform,self.expectationsqform,self.varsqform,self.GPG)' % (type)
    #    return  eval(evalstring)

    def testG(self, G1, type, altModel=None, i_exclude=None, G_exclude=None):
        """
        Params:
            G1:         SNPs to be tested
            type:       moment matching davies etc
            i_exclude:  Dummy
            G_exclude:  Dummy
        """  # this used to default to ="davies"
        self._score(G1=G1)
        pv = type.pv(self.squaredform, self.expectationsqform, self.varsqform, self.GPG)
        # stat = scoretest.scoreteststat(self.squaredform,self.varsqform)
        test = {"pv": pv, "stat": self.squaredform}
        return test


class scoretest_logit(scoretest):
    __slots__ = [
        "Y",
        "X",
        "Xdagger",
        "logreg_result",
        "logreg_mod",
        "pY",
        "stdY",
        "VX",
        "pinvVX",
    ]

    def __init__(self, Y, X=None, appendbias=False):
        ## check if is binary
        uniquey = np.unique(Y)
        if not np.sort(uniquey).tolist() == [0, 1]:
            raise Exception(
                "must use binary data in {0,1} for logit tests, found:" + str(Y)
            )

        scoretest.__init__(self, Y=Y, X=X, appendbias=appendbias)
        # from sklearn.linear_model import LogisticRegression as LR
        # logreg_sk = LR(C=200000.0)
        # logreg_sk.fit( X, Y )
        import statsmodels.api as sm

        self.logreg_mod = sm.Logit(Y[:, 0], X)
        self.logreg_result = self.logreg_mod.fit(disp=0)
        self.pY = self.logreg_result.predict(X)
        self.stdY = np.sqrt(self.pY * (1.0 - self.pY))
        self.VX = self.X * np.lib.stride_tricks.as_strided(
            (self.stdY), (self.stdY.size, self.X.shape[1]), (self.stdY.itemsize, 0)
        )
        self.pinvVX = nla.pinv(self.VX)

    def _score(self, G1):
        """
        compute the score

        Inputs:
            Bartletcorrection: refers to dividing by N-D instead of D, it is used in REML
        Outputs:
            squaredform
            expectationsqform
            varsqform
            GPG=P^1/2*K*P^1/2 (take eigenvalues of this for Davies method)
        """
        Y = self.Y
        X = self.X
        N = Y.shape[0]
        if Y.ndim == 1:
            P = 1  # num of phenotypes
        else:
            P = Y.shape[1]
        if X is None:
            D = 1  # num of covariates (and assumes they are independent)
        else:
            D = X.shape[1]
        RxY = (
            self.Y.flatten() - self.pY
        )  # residual of y regressed on X, which here, is equivalent to sigma2*Py (P is the projection matrix, which is idempotent)
        VG = G1 * np.lib.stride_tricks.as_strided(
            self.stdY, (self.stdY.size, G1.shape[1]), (self.stdY.itemsize, 0)
        )
        GY = G1.T.dot(RxY)
        squaredform = (GY * GY).sum() / (2.0 * P)

        RxVG, Xd = linreg(VG, X=self.VX, Xdagger=self.pinvVX, rcond=-1)
        if G1.shape[0] < G1.shape[1]:
            GPG = RxVG.dot(RxVG.T) / (2.0 * P)
        else:
            GPG = RxVG.T.dot(RxVG) / (2.0 * P)
        self.squaredform = squaredform
        self.expectationsqform = None
        self.varsqform = None
        self.GPG = GPG
        return squaredform, GPG


class scoretest2K(scoretest):
    __slots__ = [
        "K",
        "PxKPx",
        "G0",
        "U",
        "S",
        "Xdagger",
        "UY",
        "UUY",
        "YUUY",
        "optparams",
        "expectedinfo",
        "lowrank",
        "Neff",
    ]

    def __init__(
        self, Y, X=None, K=None, G0=None, appendbias=False, forcefullrank=False
    ):
        scoretest.__init__(self, Y=Y, X=X, appendbias=appendbias)
        self.Xdagger = None
        self.G0 = G0
        self.K = K
        # compute the spectral decomposition of K
        self.lowrank = False
        N = Y.shape[0]
        if Y.ndim == 1:
            P = 1
        else:
            P = Y.shape[1]
        D = 1
        if X is not None:
            D = X.shape[1]
        self.Neff = N - D
        if self.K is not None:
            ar = np.arange(self.K.shape[0])
            self.K[ar, ar] += 1.0
            self.PxKPx, self.Xdagger = linreg(
                Y=(self.K), X=self.X, Xdagger=self.Xdagger
            )
            self.PxKPx, self.Xdagger = linreg(
                Y=self.PxKPx.T, X=self.X, Xdagger=self.Xdagger
            )
            [self.S, self.U] = LA.eigh(self.PxKPx)
            self.K[ar, ar] -= 1.0
            self.U = self.U[:, D:N]
            self.S = self.S[D:N] - 1.0
        elif 0.7 * (self.Neff) <= self.G0.shape[1] or forcefullrank:
            self.K = self.G0.dot(self.G0.T)
            # BR: changed K to self.K (K is not defined)
            ar = np.arange(self.K.shape[0])
            self.K[ar, ar] += 1.0
            self.PxKPx, self.Xdagger = linreg(
                Y=(self.K), X=self.X, Xdagger=self.Xdagger
            )
            self.PxKPx, self.Xdagger = linreg(
                Y=self.PxKPx.T, X=self.X, Xdagger=self.Xdagger
            )
            self.K[ar, ar] -= 1.0
            # BR: changed PxKPx to self.PxKPx (PxKPx is not defined)
            [self.S, self.U] = LA.eigh(self.PxKPx)
            self.U = self.U[:, D:N]
            self.S = self.S[D:N] - 1.0
        else:
            PxG, self.Xdagger = linreg(Y=self.G0, X=self.X, Xdagger=self.Xdagger)
            [self.U, self.S, V] = LA.svd(PxG, False, True)  #!!!use big_svd?
            inonzero = self.S > 1e-10
            self.S = self.S[inonzero] * self.S[inonzero]
            self.U = self.U[:, inonzero]
            self.lowrank = True
            pass

        # rotate the phenotype as well as the fixed effects
        self.UY = self.U.T.dot(self.Y)

        if self.lowrank:
            Yres, self.Xdagger = linreg(Y=self.Y, X=self.X, Xdagger=self.Xdagger)
            self.UUY = Yres - self.U.dot(self.UY)
            self.YUUY = (self.UUY * self.UUY).sum()
            pass

        # learn null model
        resmin = [None]

        def f(x, resmin=resmin, **kwargs):
            res = self._nLLeval(h2=x)
            if (resmin[0] is None) or (res["nLL"] < resmin[0]["nLL"]):
                resmin[0] = res
            return res["nLL"]

        min = minimize1D(f, evalgrid=None, nGrid=20, minval=0.0, maxval=0.99999)

        self.optparams = resmin[0]

        # pre-compute model parameters
        self.expectedinfo = np.zeros((2, 2))
        # tr(PIPI)
        Sd = 1.0 / ((1.0 - self.optparams["h2"]) + self.optparams["h2"] * self.S)
        Sd *= Sd
        self.expectedinfo[0, 0] = (
            Sd
        ).sum()  # /(self.optparams['sigma2']*self.optparams['sigma2'])
        if self.lowrank:
            self.expectedinfo[0, 0] += ((self.Neff - self.S.shape[0])) / (
                (1.0 - self.optparams["h2"]) * (1.0 - self.optparams["h2"])
            )
        # tr(PKPI)
        Sd *= self.S
        self.expectedinfo[1, 0] = (
            Sd
        ).sum()  # /(self.optparams['sigma2']*self.optparams['sigma2'])
        self.expectedinfo[0, 1] = self.expectedinfo[1, 0]
        # tr(PKPK)
        Sd *= self.S
        self.expectedinfo[1, 1] = (
            Sd
        ).sum()  # /(self.optparams['sigma2']*self.optparams['sigma2'])
        self.expectedinfo *= (
            0.5 * P / (self.optparams["sigma2"] * self.optparams["sigma2"])
        )

        pass

    def _nLLeval(self, h2=0.0):
        """
        evaluate -ln( N( U^T*y | U^T*X*beta , h2*S + (1-h2)*I ) ),
        where K = USU^T
        --------------------------------------------------------------------------
        Input:
        h2      : mixture weight between K and Identity (environmental noise)
        --------------------------------------------------------------------------
        Output dictionary:
        'nLL'       : negative log-likelihood
        'sigma2'    : the model variance sigma^2
        'h2'        : mixture weight between Covariance and noise
        --------------------------------------------------------------------------
        """
        if (h2 < 0.0) or (h2 >= 1.0):
            return {"nLL": 3e20, "h2": h2}
        k = self.S.shape[0]
        N = self.Y.shape[0]
        if self.Y.ndim == 1:
            P = 1
        else:
            P = self.Y.shape[1]

        Sd = h2 * self.S + (1.0 - h2)
        UYS = self.UY / np.lib.stride_tricks.as_strided(
            Sd, (Sd.size, self.UY.shape[1]), (Sd.itemsize, 0)
        )

        YKY = (UYS * self.UY).sum()

        logdetK = np.log(Sd).sum()

        if self.lowrank:  # low rank part
            YKY += self.YUUY / (1.0 - h2)
            logdetK += np.log(1.0 - h2) * (self.Neff * P - k)

        sigma2 = YKY / (self.Neff * P)
        nLL = 0.5 * (logdetK + self.Neff * P * (np.log(2.0 * np.pi * sigma2) + 1))
        result = {"nLL": nLL, "sigma2": sigma2, "h2": h2}
        return result

    def _score(self, G1):
        """
        compute the score with a background kernel
        """
        # if 1:
        #    #background kernel
        #    self.K=self.G.dot(self.G.T)
        #    h2 = self.optparams['h2']
        #    sig = self.optparams['sigma2']
        #    V = h2*self.K + (1-h2)*np.eye(self.K.shape[0])
        #    V*=sig
        #    Vi=LA.inv(V)
        #    P =LA.inv(self.X.T.dot(Vi).dot(self.X))
        #    P=self.X.dot(P.dot(self.X.T))
        #    P=Vi.dot(P.dot(Vi))
        #    Px = Vi-P

        P = self.UY.shape[1]
        resG, Xdagger = linreg(Y=G1, X=self.X, Xdagger=self.Xdagger)
        sigma2e = (1.0 - self.optparams["h2"]) * self.optparams["sigma2"]
        sigma2g = self.optparams["h2"] * self.optparams["sigma2"]
        UG = self.U.T.dot(resG)
        if self.lowrank:
            UUG = resG - self.U.dot(UG)
        Sd = 1.0 / (self.S * sigma2g + sigma2e)
        SUG = UG * np.lib.stride_tricks.as_strided(
            Sd, (Sd.size, UG.shape[1]), (Sd.itemsize, 0)
        )
        # tr(YPGGPY)
        GPY = SUG.T.dot(self.UY)
        if self.lowrank:
            GPY += UUG.T.dot(self.UUY) / sigma2e
        squaredform = 0.5 * (GPY * GPY).sum()
        # tr(PGG)
        if G1.shape[0] > G1.shape[1]:
            GPG = SUG.T.dot(UG)
        else:
            GPG = SUG.dot(UG.T)
        expectationsqform = 0.5 * P * GPG.trace()
        # tr(PGGPGG)
        trPGGPGG = 0.5 * P * (GPG * GPG).sum()
        # tr(PGGPI)
        SUG *= SUG
        expectedInfoCross = np.empty(2)
        expectedInfoCross[0] = 0.5 * P * SUG.sum()
        # tr(PGGPK)
        SUG *= np.lib.stride_tricks.as_strided(
            self.S, (self.S.size, SUG.shape[1]), (self.S.itemsize, 0)
        )
        expectedInfoCross[1] = 0.5 * P * SUG.sum()
        if self.lowrank:
            if G1.shape[0] > G1.shape[1]:
                GPG_lowr = UUG.T.dot(UUG) / sigma2e
            else:
                GPG_lowr = UUG.dot(UUG.T) / sigma2e
            GPG += GPG_lowr
            # tr(PGGPGG)
            expectationsqform += 0.5 * P * GPG_lowr.trace()
            trPGGPGG += 0.5 * P * (GPG_lowr * GPG_lowr).sum()
            # tr(PGGPI)
            expectedInfoCross[0] += 0.5 * P * GPG_lowr.trace() / (sigma2e)
        varsqform = 1.0 / (
            trPGGPGG
            - expectedInfoCross.dot(LA.inv(self.expectedinfo).dot(expectedInfoCross))
        )
        self.squaredform = squaredform
        self.expectationsqform = expectationsqform
        self.varsqform = varsqform
        self.GPG = GPG * 0.5
        return self.squaredform, self.expectationsqform, self.varsqform, self.GPG

    def _findH2(self, nGridH2=10, minH2=0.0, maxH2=0.99999, **kwargs):
        """
        Find the optimal h2 for a given K.
        (default maxH2 value is set to a value smaller than 1 to avoid loss of positive definiteness of the final model covariance)
        --------------------------------------------------------------------------
        Input:
        nGridH2 : number of h2-grid points to evaluate the negative log-likelihood at
        minH2   : minimum value for h2 optimization
        maxH2   : maximum value for h2 optimization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal h2
        --------------------------------------------------------------------------
        """
        # f = lambda x : (self.nLLeval(h2=x,**kwargs)['nLL'])
        resmin = [None]

        def f(x, resmin=resmin, **kwargs):
            res = self._nLLeval(h2=x, **kwargs)
            if (resmin[0] is None) or (res["nLL"] < resmin[0]["nLL"]):
                resmin[0] = res
            return res["nLL"]

        min = minimize1D(f=f, nGrid=nGridH2, minval=minH2, maxval=maxH2)
        return resmin[0]


def linreg(Y, X=None, Xdagger=None, rcond=-1):
    if Y.ndim == 1:
        P = 1
    else:
        P = Y.shape[1]
    if X is None:
        RxY = Y - Y.mean(0)
        return RxY, None
    else:
        if Xdagger is None:
            # Xdagger = LA.pinv(X,rcond) #can be ridiculously slow (solves a linear system), 20 seconds instead of 0.1 sec.
            Xdagger = nla.pinv(X)  # SVD-based, and seems fast
        RxY = Y - X.dot(Xdagger.dot(Y))
        return RxY, Xdagger


def scoreNoK(Y, X=None, Xdagger=None, G=None, sigma2=None, Bartletcorrection=True):
    """
    compute the score

    Inputs:
        Bartletcorrection: refers to dividing by N-D instead of D, it is used in REML
    Outputs:
        squaredform
        expectationsqform
        varsqform
        GPG=P^1/2*K*P^1/2 (take eigenvalues of this for Davies method)
    """
    N = Y.shape[0]
    if Y.ndim == 1:
        P = 1  # num of phenotypes
    else:
        P = Y.shape[1]
    if X is None:
        D = 1  # num of covariates (and assumes they are independent)
    else:
        D = X.shape[1]
    RxY, Xdagger = linreg(
        Y=Y, X=X, Xdagger=Xdagger
    )  # residual of y regressed on X, which here, is equivalent to sigma2*Py (P is the projection matrix, which is idempotent)

    if (
        sigma2 is None
    ):  #     note: Xdagger is pseudo inverse of X, or (X^T*X)^1*X^T such that Xdagger*y=beta
        if Bartletcorrection:
            sigma2 = (RxY * RxY).sum() / ((N - D) * P)
        else:
            sigma2 = (RxY * RxY).sum() / (N * P)

    RxG, Xdagger = linreg(
        Y=G, X=X, Xdagger=Xdagger
    )  # residual of G regressed on X, which here, is equivalent to PG (P is the projection matrix, and in this one kernel case, is idempotent)
    #     note: P is never computed explicitly, only via residuals such as Py=1/sigma2(I-Xdagger*X)y and PG=1/sigma2(I-Xdagger*X)G
    #     also note that "RxY"=Py=1/sigma2*(I-Xdagger*X)y is nothing more (except for 1/sigma2) than the residual of y regressed on X (i.e. y-X*beta),
    #     and similarly for PG="RxG"
    GtRxY = G.T.dot(RxY)
    squaredform = ((GtRxY * GtRxY).sum()) * (
        0.5 / (sigma2 * sigma2)
    )  # yPKPy=yPG^T*GPy=(yPG^T)*(yPG^T)^T
    if G.shape[0] > G.shape[1]:
        GPG = np.dot(
            RxG.T, RxG
        )  # GPG is always a square matrix in the smaller dimension
    else:
        GPG = np.dot(RxG, RxG.T)
    expectationsqform = (
        P * (GPG.trace()) * (0.5 / sigma2)
    )  # note this is Trace(PKP)=Trace(PPK)=Trace(PK), for P=projection matrix in comment, and in the code P=1=#phen
    expectedinfo00 = P * (GPG * GPG).sum() * (0.5 / (sigma2 * sigma2))
    expectedinfo10 = expectationsqform / sigma2  # P*0.5/(sigma2*sigma2)*GPG.trace()
    expectedinfo11 = P * (N - D) * (0.5 / (sigma2 * sigma2))
    varsqform = 1.0 / (
        expectedinfo00 - expectedinfo10 * expectedinfo10 / expectedinfo11
    )
    # if 1:
    #    XXi=LA.inv(X.T.dot(X))
    #    Px=(np.eye(N)-X.dot(XXi).dot(X.T))/sigma2
    # pdb.set_trace()
    GPG /= (
        sigma2 * 2.0
    )  # what we will take eigenvalues of for Davies (which is P^1/2*K*P^1/2)

    # for debugging, explicitly compute GPG=P^1/2 * K * P^1/2
    # SigInv=(1/sigma2)*np.eye(N,N)
    # Phat=X.dot(LA.inv(X.T.dot(SigInv).dot(X))).dot(X.T).dot(SigInv)
    # PP=SigInv.dot(np.eye(N,N)-Phat)
    # K=G.dot(G.T)
    # PKP=PP.dot(K).dot(PP)
    # ss.stats(PKP-PKP.T)
    ##eigvalsFull=LA.eigh(PKP,eigvals_only=True)
    # eigvalsFull2=LA.eigvals(PKP)
    # eigvalsLow =LA.eigh(GPG,eigvals_only=True)
    # GPG=PKP*0.5
    # pdb.set_trace()

    return squaredform, expectationsqform, varsqform, GPG, GtRxY * (0.25 / sigma2)


if __name__ == "__main__":
    if 1:  # example p-value computation for sample data
        # specify the directory that contains the data
        datadir = "data"  # os.path.join('twokerneltest','data')
        # specify the directory that contains the alternative models in form of ped files
        datadiralt = os.path.join(datadir, "altmodels")
        pedfilesalt = glob.glob(os.path.join(datadiralt, "*.ped"))
        for i in range(len(pedfilesalt)):
            pedfilesalt[i] = pedfilesalt[i][0:-4]

        phenofile = os.path.join(datadir, "phen.N1000.M5000.txt")
        covarfile = os.path.join(datadir, "covariates.N1000.M5000.txt")
        # base0 = os.path.join(datadir,'snpDat.N1000.M5000.20Snps')
        base0 = os.path.join(datadir, "snpDat.N1000.M5000.10_20Snps")

        # specify index of the phenotype to be tested
        ipheno = 0  # only one phenotype in file, use the first one

        # exclusion parameters (correction for proximal contamination)
        mindist = (
            10  # minimum distance to alternative SNPs to be included in null model
        )
        idist = 2  # use genetic distance
        # idist = 3      #use basepair distance

        # run the example
        logging.info(("\n\n\nrunning real data example"))
        logging.info(("base file of null model: %s" % base0))
        logging.info(("testing all SNP sets in %s" % datadiralt))
        result = testPedfilesFromDir(
            phenofile,
            base0,
            pedfilesalt,
            ipheno=ipheno,
            mindist=mindist,
            idist=idist,
            covarfile=covarfile,
        )



# flake8: noqa: E501

import sys
import numpy as np
import logging
import unittest
import os.path
import doctest
import pandas as pd
from numpy.random import RandomState
from pathlib import Path

from fastlmm.association import single_snp
from fastlmm.association import single_snp_linreg
import pysnptools.util.pheno as pstpheno
from fastlmm.feature_selection.test import TestFeatureSelection
from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc
from pysnptools.kernelreader import Identity as KernelIdentity
from pysnptools.standardizer import Unit
from pysnptools.snpreader import Bed, Pheno, SnpData
from pysnptools.kernelreader import SnpKernel
from unittest.mock import patch


class TestSingleSnp(unittest.TestCase):
    # !!! created a Expect Durbin, too

    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "..")
        )
        self.bedbase = os.path.join(
            self.pythonpath, "tests/datasets/all_chr.maf0.001.N300"
        )
        self.phen_fn = os.path.join(
            self.pythonpath, "tests/datasets/phenSynthFrom22.23.N300.randcidorder.txt"
        )
        self.cov_fn = os.path.join(
            self.pythonpath, "tests/datasets/all_chr.maf0.001.covariates.N300.txt"
        )

    tempout_dir = "tempout/single_snp"

    def test_match_cpp(self):
        """
        match
            FaSTLMM.207\Data\DemoData>..\.cd.\bin\windows\cpp_mkl\fastlmmc -bfile snps -extract topsnps.txt -bfileSim snps -extractSim ASout.snps.txt -pheno pheno.txt -covar covariate.txt -out topsnps.singlesnp.txt -logDelta 0 -verbose 100

        """
        logging.info("TestSingleSnp test_match_cpp")
        snps = Bed(
            os.path.join(self.pythonpath, "tests/datasets/selecttest/snps.bed"),
            count_A1=False,
        )
        pheno = os.path.join(self.pythonpath, "tests/datasets/selecttest/pheno.txt")
        covar = os.path.join(self.pythonpath, "tests/datasets/selecttest/covariate.txt")
        sim_sid = [
            "snp26250_m0_.19m1_.19",
            "snp82500_m0_.28m1_.28",
            "snp63751_m0_.23m1_.23",
            "snp48753_m0_.4m1_.4",
            "snp45001_m0_.26m1_.26",
            "snp52500_m0_.05m1_.05",
            "snp75002_m0_.39m1_.39",
            "snp41253_m0_.07m1_.07",
            "snp11253_m0_.2m1_.2",
            "snp86250_m0_.33m1_.33",
            "snp3753_m0_.23m1_.23",
            "snp75003_m0_.32m1_.32",
            "snp30002_m0_.25m1_.25",
            "snp26252_m0_.19m1_.19",
            "snp67501_m0_.15m1_.15",
            "snp63750_m0_.28m1_.28",
            "snp30001_m0_.28m1_.28",
            "snp52502_m0_.35m1_.35",
            "snp33752_m0_.31m1_.31",
            "snp37503_m0_.37m1_.37",
            "snp15002_m0_.11m1_.11",
            "snp3751_m0_.34m1_.34",
            "snp7502_m0_.18m1_.18",
            "snp52503_m0_.3m1_.3",
            "snp30000_m0_.39m1_.39",
            "isnp4457_m0_.11m1_.11",
            "isnp23145_m0_.2m1_.2",
            "snp60001_m0_.39m1_.39",
            "snp33753_m0_.16m1_.16",
            "isnp60813_m0_.2m1_.2",
            "snp82502_m0_.34m1_.34",
            "snp11252_m0_.13m1_.13",
        ]
        sim_idx = snps.sid_to_index(sim_sid)
        test_sid = [
            "snp26250_m0_.19m1_.19",
            "snp63751_m0_.23m1_.23",
            "snp82500_m0_.28m1_.28",
            "snp48753_m0_.4m1_.4",
            "snp45001_m0_.26m1_.26",
            "snp52500_m0_.05m1_.05",
            "snp75002_m0_.39m1_.39",
            "snp41253_m0_.07m1_.07",
            "snp86250_m0_.33m1_.33",
            "snp15002_m0_.11m1_.11",
            "snp33752_m0_.31m1_.31",
            "snp26252_m0_.19m1_.19",
            "snp30001_m0_.28m1_.28",
            "snp11253_m0_.2m1_.2",
            "snp67501_m0_.15m1_.15",
            "snp3753_m0_.23m1_.23",
            "snp52502_m0_.35m1_.35",
            "snp30000_m0_.39m1_.39",
            "snp30002_m0_.25m1_.25",
        ]
        test_idx = snps.sid_to_index(test_sid)

        for G0, G1 in [
            (snps[:, sim_idx], KernelIdentity(snps.iid)),
            (KernelIdentity(snps.iid), snps[:, sim_idx]),
        ]:
            frame_h2 = single_snp(
                test_snps=snps[:, test_idx],
                pheno=pheno,
                G0=G0,
                G1=G1,
                covar=covar,
                h2=0.5,
                leave_out_one_chrom=False,
                count_A1=False,
            )
            frame_log_delta = single_snp(
                test_snps=snps[:, test_idx],
                pheno=pheno,
                G0=G0,
                G1=G1,
                covar=covar,
                log_delta=0,
                leave_out_one_chrom=False,
                count_A1=False,
            )
            for frame in [frame_h2, frame_log_delta]:
                referenceOutfile = TestFeatureSelection.reference_file(
                    "single_snp/topsnps.single.txt"
                )
                reference = pd.read_csv(
                    referenceOutfile, sep="\t"
                )  # We've manually remove all comments and blank lines from this file
                assert len(frame) == len(reference)
                for _, row in reference.iterrows():
                    sid = row.SNP
                    pvalue = frame[frame["SNP"] == sid].iloc[0].PValue
                    reldiff = abs(row.Pvalue - pvalue) / row.Pvalue
                    assert (
                        reldiff < 0.035
                    ), "'{0}' pvalue_list differ too much {4} -- {2} vs {3}".format(
                        sid, None, row.Pvalue, pvalue, reldiff
                    )

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name + ".txt")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def test_mixing(self):
        logging.info("TestSingleSnp test_mixing")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file_name = self.file_name("mixing")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps[:, 10:100],
            leave_out_one_chrom=False,
            covar=covar,
            G1=test_snps[:, 100:200],
            mixing=None,
            output_file_name=output_file_name,
            count_A1=False,
        )

        self.compare_files(frame, "mixing")

    def test_mixingKs(self):
        logging.info("TestSingleSnp test_mixingKs")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file_name = self.file_name("mixingKs")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            K0=SnpKernel(test_snps[:, 10:100], Unit()),
            leave_out_one_chrom=False,
            covar=covar,
            K1=SnpKernel(test_snps[:, 100:200], Unit()),
            mixing=None,
            output_file_name=output_file_name,
            count_A1=False,
        )

        self.compare_files(frame, "mixing")

    def test_mixid(self):
        logging.info("TestSingleSnp test_mixid")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file_name = self.file_name("mixid")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps[:, 10:100],
            leave_out_one_chrom=False,
            covar=covar,
            K1=KernelIdentity(test_snps.iid),
            mixing=0.25,
            output_file_name=output_file_name,
            count_A1=False,
        )

        self.compare_files(frame, "mixid")

    def test_one(self):
        logging.info("TestSingleSnp test_one")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("one")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            mixing=0,
            leave_out_one_chrom=False,
            G0=test_snps,
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "one")

    def test_snp_fract_var_exp(self):
        from fastlmm.util import example_file  # Download and return local file name

        logging.info("TestSingleSnp test_snp_fract_var_exp")
        bed_fn = example_file("tests/datasets/synth/all.*", "*.bed")
        pheno_fn = example_file("tests/datasets/synth/pheno_10_causals.txt")
        cov_fn = example_file("tests/datasets/synth/cov.txt")

        output_file = self.file_name("snp_fract_var_exp")
        frame = single_snp(
            bed_fn,
            pheno_fn,
            covar=cov_fn,
            show_snp_fract_var_exp=True,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(
            frame,
            "snp_fract_var_exp",
            columns=["PValue", "EffectSize", "SnpFractVarExpl"],
        )

    def test_zero_pheno(self):
        logging.info("TestSingleSnp test_zero_pheno")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = Pheno(self.phen_fn)[:, 0:0]
        covar = self.cov_fn

        got_expected_fail = False
        try:
            frame = single_snp(
                test_snps=test_snps[:, :10],
                pheno=pheno,
                mixing=0,
                leave_out_one_chrom=False,
                G0=test_snps,
                covar=covar,
                count_A1=False,
            )
        except Exception as e:
            got_expected_fail = True
        assert got_expected_fail, "Did not get expected fail"

    def test_missing_covar(self):
        logging.info("TestSingleSnp test_missing_covar")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = Pheno(self.cov_fn).read()
        covar.val[0, 0] = np.nan

        got_expected_fail = False
        try:
            frame = single_snp(
                test_snps=test_snps[:, :10],
                pheno=pheno,
                mixing=0,
                leave_out_one_chrom=False,
                G0=test_snps,
                covar=covar,
                count_A1=False,
            )
        except Exception as e:
            got_expected_fail = True
        assert got_expected_fail, "Did not get expected fail"

        covar_by_chrom = {chrom: covar for chrom in set(test_snps.pos[:, 0])}
        got_expected_fail = False
        try:
            frame = single_snp(
                test_snps=test_snps[:, :10],
                pheno=pheno,
                mixing=0,
                leave_out_one_chrom=True,
                G0=test_snps,
                covar_by_chrom=covar_by_chrom,
                count_A1=False,
            )
        except Exception as e:
            got_expected_fail = True
        assert got_expected_fail, "Did not get expected fail"

    def test_thres(self):
        logging.info("TestSingleSnp test_thres")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        reffile = TestFeatureSelection.reference_file("single_snp/one.txt")
        reference = pd.read_csv(reffile, delimiter="\s", comment=None, engine="python")

        for random_seed in [0, 1]:
            for pvalue_threshold in [0.5, None, 1.0]:
                for random_threshold in [0.5, None, 1.0]:
                    frame = single_snp(
                        test_snps=test_snps[:, :10],
                        pheno=pheno,
                        mixing=0,
                        leave_out_one_chrom=False,
                        G0=test_snps,
                        covar=covar,
                        pvalue_threshold=pvalue_threshold,
                        random_threshold=random_threshold,
                        random_seed=random_seed,
                        count_A1=False,
                    )

                    assert len(frame) <= len(
                        reference
                    ), "# of pairs differs from file '{0}'".format(reffile)
                    if len(frame) < len(reference):
                        assert frame["PValueCount"].iloc[0] == len(
                            reference
                        ), "row_count doesn't match the reference # of rows"
                    if random_threshold is not None:
                        assert np.all(
                            frame["RandomThreshold"] == random_threshold
                        ), "Expect all rows to have 'RandomThreshold'"
                        assert np.all(
                            frame["RandomSeed"] == random_seed
                        ), "Expect all rows to have right 'RandomSeed'"
                        if pvalue_threshold is not None:
                            assert np.all(
                                (frame["RandomValue"] <= random_threshold)
                                + (frame["PValue"] <= pvalue_threshold)
                            ), "Expect all rows have random value or pvalue less than threshold"
                    if pvalue_threshold is not None:
                        assert np.all(
                            frame["PValueThreshold"] == pvalue_threshold
                        ), "Expect all rows to have 'PValueThreshold'"
                    for _, row in frame.iterrows():
                        sid = row.SNP
                        pvalue = reference[reference["SNP"] == sid].iloc[0].PValue
                        diff = abs(row.PValue - pvalue)
                        if diff > 1e-5 or np.isnan(diff):
                            raise Exception(
                                "pair {0} differs too much from file '{1}'".format(
                                    sid, reffile
                                )
                            )
                        assert abs(row.PValue - pvalue) < 1e-5, "wrong"

    def test_linreg(self):
        logging.info("TestSingleSnp test_linreg")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("linreg")

        frame1 = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            mixing=0,
            leave_out_one_chrom=False,
            G0=KernelIdentity(iid=test_snps.iid),
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )

        frame1 = frame1[["sid_index", "SNP", "Chr", "GenDist", "ChrPos", "PValue"]]
        self.compare_files(frame1, "linreg")

        with patch.dict("os.environ", {"ARRAY_MODULE": "numpy"}) as _:
            frame2 = single_snp_linreg(
                test_snps=test_snps[:, :10],
                pheno=pheno,
                covar=covar,
                output_file_name=output_file,
            )
        self.compare_files(frame2, "linreg")

    def test_noK0(self):
        logging.info("TestSingleSnp test_noK0")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("noK0")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            mixing=1,
            leave_out_one_chrom=False,
            G1=test_snps,
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "one")

    def test_gb_goal(self):
        logging.info("TestSingleSnp test_gb_goal")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("gb_goal")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            mixing=0,
            leave_out_one_chrom=False,
            G0=test_snps,
            covar=covar,
            GB_goal=0,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "one")

        output_file = self.file_name("gb_goal2")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            mixing=0,
            leave_out_one_chrom=False,
            G0=test_snps,
            covar=covar,
            GB_goal=0.12,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "one")

    def test_other(self):
        logging.info("TestSingleSnp test_other")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("other")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            leave_out_one_chrom=False,
            K1=test_snps,
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "one")

    def test_none(self):
        logging.info("TestSingleSnp test_none")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("none")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            mixing=0,
            leave_out_one_chrom=False,
            K0=KernelIdentity(test_snps.iid),
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "none")

    def test_interact(self):
        logging.info("TestSingleSnp test_interact")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("interact")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            mixing=0,
            leave_out_one_chrom=False,
            G0=test_snps,
            covar=covar,
            interact_with_snp=1,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "interact")

    def test_preload_files(self):
        logging.info("TestSingleSnp test_preload_files")
        test_snps = self.bedbase
        pheno = pstpheno.loadOnePhen(self.phen_fn, vectorize=True)
        covar = pstpheno.loadPhen(self.cov_fn)
        bed = Bed(test_snps, count_A1=False)

        output_file_name = self.file_name("preload_files")

        frame = single_snp(
            test_snps=bed[:, :10],
            pheno=pheno,
            G0=test_snps,
            mixing=0,
            leave_out_one_chrom=False,
            covar=covar,
            output_file_name=output_file_name,
            count_A1=False,
        )
        self.compare_files(frame, "one")

    def test_SNC(self):
        logging.info("TestSNC")
        test_snps = self.bedbase
        pheno = pstpheno.loadOnePhen(self.phen_fn, vectorize=True)
        covar = pstpheno.loadPhen(self.cov_fn)
        bed = Bed(test_snps, count_A1=False)
        snc = bed.read()
        snc.val[:, 2] = 0  # make SNP #2 have constant values (aka a SNC)

        output_file_name = self.file_name("snc")

        frame = single_snp(
            test_snps=snc[:, :10],
            pheno=pheno,
            G0=snc,
            mixing=0,
            leave_out_one_chrom=False,
            covar=covar,
            output_file_name=output_file_name,
            count_A1=False,
        )
        self.compare_files(frame, "snc")

    def test_G0_has_reader(self):
        logging.info("TestSingleSnp test_G0_has_reader")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file_name = self.file_name("G0_has_reader")

        frame0 = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps,
            leave_out_one_chrom=False,
            covar=covar,
            mixing=0,
            output_file_name=output_file_name,
            count_A1=False,
        )
        self.compare_files(frame0, "one")

        frame1 = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=KernelIdentity(test_snps.iid),
            G1=test_snps,
            leave_out_one_chrom=False,
            covar=covar,
            mixing=1,
            output_file_name=output_file_name,
            count_A1=False,
        )
        self.compare_files(frame1, "one")

    def test_no_cov(self):
        logging.info("TestSingleSnp test_no_cov")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn

        output_file_name = self.file_name("no_cov")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps,
            mixing=0,
            leave_out_one_chrom=False,
            output_file_name=output_file_name,
            count_A1=False,
        )

        self.compare_files(frame, "no_cov")

    def test_no_cov_b(self):
        logging.info("TestSingleSnp test_no_cov_b")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn

        output_file_name = self.file_name("no_cov_b")
        covar = pstpheno.loadPhen(self.cov_fn)
        covar["vals"] = np.delete(covar["vals"], np.s_[:], 1)  # Remove all the columns
        covar["header"] = []

        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps,
            leave_out_one_chrom=False,
            covar=covar,
            mixing=0,
            output_file_name=output_file_name,
            count_A1=False,
        )

        self.compare_files(frame, "no_cov")

    def test_G1(self):
        logging.info("TestSingleSnp test_G1")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file_name = self.file_name("G1")
        for force_full_rank, force_low_rank in [
            (False, True),
            (False, False),
            (True, False),
        ]:
            logging.info("{0},{1}".format(force_full_rank, force_low_rank))
            frame = single_snp(
                test_snps=test_snps[:, :10],
                pheno=pheno,
                G0=test_snps[:, 10:100],
                leave_out_one_chrom=False,
                covar=covar,
                G1=test_snps[:, 100:200],
                mixing=0.5,
                force_full_rank=force_full_rank,
                force_low_rank=force_low_rank,
                output_file_name=output_file_name,
                count_A1=False,
            )
            self.compare_files(frame, "G1")

    def test_file_cache(self):
        logging.info("TestSingleSnp test_file_cache")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file_name = self.file_name("G1")
        cache_file = self.file_name("cache_file") + ".npz"
        if os.path.exists(cache_file):
            os.remove(cache_file)
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps[:, 10:100],
            leave_out_one_chrom=False,
            covar=covar,
            G1=test_snps[:, 100:200],
            mixing=0.5,
            output_file_name=output_file_name,
            cache_file=cache_file,
            count_A1=False,
        )
        self.compare_files(frame, "G1")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps[:, 10:100],
            leave_out_one_chrom=False,
            covar=covar,
            G1=test_snps[:, 100:200],
            mixing=0.5,
            output_file_name=output_file_name,
            cache_file=cache_file,
            count_A1=False,
        )
        self.compare_files(frame, "G1")

    def test_G1_mixing(self):
        logging.info("TestSingleSnp test_G1_mixing")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file_name = self.file_name("G1_mixing")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps,
            leave_out_one_chrom=False,
            covar=covar,
            G1=test_snps[:, 100:200],
            mixing=0,
            output_file_name=output_file_name,
            count_A1=False,
        )

        self.compare_files(frame, "one")

    def test_unknown_sid(self):
        logging.info("TestSingleSnp test_unknown_sid")

        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        try:
            frame = single_snp(
                test_snps=test_snps,
                G0=test_snps,
                pheno=pheno,
                leave_out_one_chrom=False,
                mixing=0,
                covar=covar,
                sid_list=["1_4", "bogus sid", "1_9"],
                count_A1=False,
            )
            failed = False
        except:
            failed = True

        assert failed

    def test_cid_intersect(self):
        logging.info("TestSingleSnp test_cid_intersect")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = pstpheno.loadOnePhen(self.phen_fn, vectorize=True)
        pheno["iid"] = np.vstack([pheno["iid"][::-1], [["Bogus", "Bogus"]]])
        pheno["vals"] = np.hstack([pheno["vals"][::-1], [-34343]])

        covar = self.cov_fn
        output_file_name = self.file_name("cid_intersect")
        frame = single_snp(
            test_snps=test_snps[:, :10],
            pheno=pheno,
            G0=test_snps,
            leave_out_one_chrom=False,
            covar=covar,
            mixing=0,
            output_file_name=output_file_name,
            count_A1=False,
        )

        self.compare_files(frame, "one")

    def compare_files(self, frame, ref_base, columns=["PValue"]):
        reffile = TestFeatureSelection.reference_file("single_snp/" + ref_base + ".txt")

        # sid_list,pvalue_list = frame['SNP'].values,frame['Pvalue'].values

        # sid_to_pvalue = {}
        # for index, sid in enumerate(sid_list):
        #    sid_to_pvalue[sid] = pvalue_list[index]

        reference = pd.read_csv(reffile, delimiter="\s", comment=None, engine="python")
        assert len(frame) == len(
            reference
        ), "# of pairs differs from file '{0}'".format(reffile)
        for _, row in reference.iterrows():
            sid = row.SNP
            for column in columns:
                value = frame[frame["SNP"] == sid].iloc[0][column]
                diff = abs(row[column] - value)
                if diff > 1e-5 or np.isnan(diff):
                    raise Exception(
                        f"pair {sid} differs too much on {column} from file '{reffile}'"
                    )
                assert abs(row[column] - value) < 1e-5, "wrong"

    def test_doctest(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        result = doctest.testmod(sys.modules["fastlmm.association.single_snp"])
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


class TestSingleSnpLeaveOutOneChrom(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "..")
        )
        self.bedbase = os.path.join(
            self.pythonpath, "fastlmm/feature_selection/examples/toydata.5chrom.bed"
        )
        self.phen_fn = os.path.join(
            self.pythonpath, "fastlmm/feature_selection/examples/toydata.phe"
        )
        self.cov_fn = os.path.join(
            self.pythonpath, "fastlmm/feature_selection/examples/toydata.cov"
        )

    tempout_dir = "tempout/single_snp"

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name + ".txt")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def test_leave_one_out_with_prekernels(self):
        logging.info("TestSingleSnpLeaveOutOneChrom test_leave_one_out_with_prekernels")
        from pysnptools.kernelstandardizer import DiagKtoN

        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        chrom_to_kernel = {}
        with patch.dict("os.environ", {"ARRAY_MODULE": "numpy"}) as _:
            for chrom in np.unique(test_snps.pos[:, 0]):
                other_snps = test_snps[:, test_snps.pos[:, 0] != chrom]
                kernel = other_snps.read_kernel(
                    standardizer=Unit(), block_size=500
                )  # Create a kernel from the SNPs not used in testing
                chrom_to_kernel[chrom] = kernel.standardize(
                    DiagKtoN()
                )  # improves the kernel numerically by making its diagonal sum to iid_count

        output_file = self.file_name("one_looc_prekernel")
        frame = single_snp(
            test_snps,
            pheno,
            covar=covar,
            K0=chrom_to_kernel,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "one_looc")

    def test_one_looc(self):
        logging.info("TestSingleSnpLeaveOutOneChrom test_one_looc")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("one_looc")
        frame = single_snp(
            test_snps,
            pheno,
            covar=covar,
            mixing=0,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "one_looc")

    def test_runner(self):
        logging.info("TestRunner")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        runner = LocalMultiProc(6, just_one_process=True)
        for map_reduce_outer in [True, False]:
            frame = single_snp(
                test_snps,
                pheno,
                covar=covar,
                mixing=0,
                count_A1=False,
                runner=runner,
                map_reduce_outer=map_reduce_outer,
            )

            self.compare_files(frame, "one_looc")

    def test_multipheno(self):
        logging.info("TestSingleSnpLeaveOutOneChrom test_multipheno")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        pheno2 = Pheno(pheno).read()
        pheno2.val[0, 0] = 100
        pheno2.val[1, 0] = -100

        cache_file = None

        if True:
            pheno12 = SnpData(
                iid=pheno2.iid,
                sid=["pheno1", "pheno2"],
                val=np.c_[Pheno(pheno).read().val, pheno2.val],
            )
            output_file = self.file_name("multipheno12")
            frame = single_snp(
                test_snps[:, ::10],
                pheno12,
                force_full_rank=True,
                covar=covar,
                cache_file=cache_file,
                output_file_name=output_file,
            )
            frame1 = frame[frame["Pheno"] == "pheno1"]
            del frame1["Pheno"]
            frame1_ef = frame1.copy()
            self.compare_files(frame1, "two_looc")
            self.compare_files_effect_size(frame1_ef, "two_looc")

            frame2 = frame[frame["Pheno"] == "pheno2"]
            del frame2["Pheno"]
            frame2_ef = frame2.copy()
            self.compare_files(frame2, "multipheno2")
            self.compare_files_effect_size(frame2_ef, "multipheno2")

        if True:
            pheno11 = SnpData(
                iid=pheno2.iid,
                sid=["pheno1a", "pheno1b"],
                val=np.c_[Pheno(pheno).read().val, Pheno(pheno).read().val],
            )
            output_file = self.file_name("multipheno11")
            frame = single_snp(
                test_snps[:, ::10],
                pheno11,
                force_full_rank=True,
                covar=covar,
                output_file_name=output_file,
                count_A1=False,
            )

            frame1 = frame[frame["Pheno"] == "pheno1a"]
            del frame1["Pheno"]
            self.compare_files(frame1, "two_looc")

            frame2 = frame[frame["Pheno"] == "pheno1b"]
            del frame2["Pheno"]
            self.compare_files(frame2, "two_looc")

        if True:
            output_file = self.file_name("multipheno1")
            frame = single_snp(
                test_snps[:, ::10],
                pheno,
                force_full_rank=True,
                covar=covar,
                output_file_name=output_file,
                count_A1=False,
            )

            self.compare_files(frame, "two_looc")

        if True:
            output_file = self.file_name("multipheno2")
            frame = single_snp(
                test_snps[:, ::10],
                pheno2,
                force_full_rank=True,
                covar=covar,
                output_file_name=output_file,
                count_A1=False,
            )

            self.compare_files(frame, "multipheno2")

        if True:
            pheno22 = SnpData(
                iid=pheno2.iid,
                sid=["pheno2a", "pheno2b"],
                val=np.c_[pheno2.val, pheno2.val],
            )
            output_file = self.file_name("multipheno22")
            frame = single_snp(
                test_snps[:, ::10],
                pheno22,
                force_full_rank=True,
                covar=covar,
                output_file_name=output_file,
                count_A1=False,
            )
            frame1 = frame[frame["Pheno"] == "pheno2a"]
            del frame1["Pheno"]
            self.compare_files(frame1, "multipheno2")

            frame2 = frame[frame["Pheno"] == "pheno2b"]
            del frame2["Pheno"]
            self.compare_files(frame2, "multipheno2")

    def test_multipheno2(self):
        logging.info("test_multipheno")
        from fastlmm.util import example_file  # Download and return local file name

        bed = Bed(example_file("tests/datasets/synth/all.*", "*.bed"), count_A1=True)[
            :, ::10
        ]
        phen_fn = example_file("tests/datasets/synth/pheno_10_causals.txt")
        cov_fn = example_file("tests/datasets/synth/cov.txt")

        random_state = RandomState(29921)
        pheno_reference = Pheno(phen_fn).read()
        for pheno_count in [2, 5, 1]:
            val = random_state.normal(
                loc=pheno_count,
                scale=pheno_count,
                size=(pheno_reference.iid_count, pheno_count),
            )
            pheno_col = ["pheno{0}".format(i) for i in range(pheno_count)]
            pheno_multi = SnpData(iid=pheno_reference.iid, sid=pheno_col, val=val)

            reference = pd.concat(
                [
                    single_snp(
                        test_snps=bed, pheno=pheno_multi[:, pheno_index], covar=cov_fn
                    )
                    for pheno_index in range(pheno_count)
                ]
            )

            for force_full_rank, force_low_rank in [
                (True, False),
                (False, True),
                (False, False),
            ]:
                frame = single_snp(
                    test_snps=bed,
                    pheno=pheno_multi,
                    covar=cov_fn,
                    force_full_rank=force_full_rank,
                    force_low_rank=force_low_rank,
                )

                assert len(frame) == len(reference), "# of pairs differs"
                for sid in sorted(
                    set(reference.SNP)
                ):  # This ignores which pheno produces which pvalue
                    pvalue_frame = np.array(sorted(frame[frame["SNP"] == sid].PValue))
                    pvalue_reference = np.array(
                        sorted(reference[reference["SNP"] == sid].PValue)
                    )
                    assert (
                        abs(pvalue_frame - pvalue_reference) < 1e-5
                    ).all, "pair {0} differs too much from reference".format(sid)

    def create_phen3(self, phen):
        from fastlmm.util import example_file  # Download and return local file name

        phen = phen.read()
        rng = np.random.RandomState(seed=0)
        val1 = phen.val.copy()
        rng.shuffle(val1)
        val2 = phen.val.copy()
        rng.shuffle(val2)
        phen3 = SnpData(
            iid=phen.iid,
            sid=["phen0", "phen1", "phen2"],
            val=np.c_[phen.val, val1, val2],
        )
        return phen3

    def test_multipheno3(self):
        from pysnptools.kernelreader import SnpKernel
        from fastlmm.util import example_file  # Download and return local file name
        from pysnptools.standardizer import Standardizer, Unit

        bed = Bed(example_file("tests/datasets/synth/all.*", "*.bed"), count_A1=True)[
            :, ::10
        ]
        phen3 = self.create_phen3(
            Pheno(example_file("tests/datasets/synth/pheno_10_causals.txt"))
        )

        combo_index = 0
        for covar, interact in [
            (None, None),
            (Pheno(example_file("tests/datasets/synth/cov.txt")), None),
            (Pheno(example_file("tests/datasets/synth/cov.txt")), 0),
        ]:
            for force_full_rank, force_low_rank in [
                (True, False),
                (False, True),
                (False, False),
            ]:
                for k0_as_snps in [True, False]:
                    logging.info(f"combo_index {combo_index}")
                    combo_index += 1
                    k0 = bed
                    if not k0_as_snps:
                        k0 = SnpKernel(k0, standardizer=Unit())

                    for leave_out_one_chrom in [False, True]:
                        logging.info(
                            [
                                covar,
                                interact,
                                force_full_rank,
                                force_low_rank,
                                k0_as_snps,
                                leave_out_one_chrom,
                            ]
                        )

                        result3 = single_snp(
                            test_snps=bed,
                            pheno=phen3,
                            covar=covar,
                            K0=k0,
                            interact_with_snp=interact,
                            force_full_rank=force_full_rank,
                            force_low_rank=force_low_rank,
                            leave_out_one_chrom=leave_out_one_chrom,
                        )

                        result_list = [
                            single_snp(
                                test_snps=bed,
                                pheno=phen3[:, i],
                                covar=covar,
                                K0=k0,
                                interact_with_snp=interact,
                                force_full_rank=force_full_rank,
                                force_low_rank=force_low_rank,
                                leave_out_one_chrom=leave_out_one_chrom,
                            )
                            for i in range(3)
                        ]

                        for i in range(3):
                            self.compare_df(
                                result3[result3["Pheno"] == phen3.sid[i]],
                                result_list[i],
                                "test_multipheno3",
                            )

    def test_multipheno_expected_exceptions(self):
        from pysnptools.kernelreader import SnpKernel
        from fastlmm.util import example_file  # Download and return local file name
        from pysnptools.standardizer import Standardizer, Unit

        bed = Bed(example_file("tests/datasets/synth/all.*", "*.bed"), count_A1=True)[
            :, ::10
        ]
        phen = Pheno(example_file("tests/datasets/synth/pheno_10_causals.txt"))
        phen3 = self.create_phen3(phen)

        got_expected_fail = False
        try:
            single_snp(
                test_snps=bed,
                pheno=phen3,
                covar=None,
                K0=bed,
                K1=bed,
                interact_with_snp=None,
                force_full_rank=False,
                force_low_rank=False,
            )
        except Exception as e:
            assert "2nd kernel" in str(e)
            got_expected_fail = True
        assert got_expected_fail, "Did not get expected fail"

        phen3.val[1, :] = np.nan  # Add a missing value to all phenos
        single_snp(
            test_snps=bed,
            pheno=phen3,
            covar=None,
            K0=bed,
            interact_with_snp=None,
            force_full_rank=False,
            force_low_rank=False,
        )
        phen3.val[0, 0] = np.nan  # Add a missing value to one pheno, but not the others
        got_expected_fail = False
        try:
            single_snp(
                test_snps=bed,
                pheno=phen3,
                covar=None,
                K0=bed,
                interact_with_snp=None,
                force_full_rank=False,
                force_low_rank=False,
            )
        except Exception as e:
            assert "multiple phenotypes" in str(e)
            got_expected_fail = True
        assert got_expected_fail, "Did not get expected fail"

    def test_cache(self):
        test_snpsx = Bed(self.bedbase, count_A1=False)
        phen1 = self.phen_fn
        phen3 = self.create_phen3(Pheno(self.phen_fn))
        covar = self.cov_fn

        for leave_out_one_chrom, ref_file1, ref_file3, test_snps in [
            (True, "one_looc", "one_looc3", test_snpsx),
            (False, "one", "one3", test_snpsx[:, :10]),
        ]:
            for force_full_rank, force_low_rank in [
                (False, True),
                (True, False),
                (False, False),
            ]:
                for pheno, ref_file in [(phen3, ref_file3), (phen1, ref_file1)]:
                    output_file = self.file_name(
                        f"cache{leave_out_one_chrom}{force_full_rank}{force_low_rank}"
                    )
                    cache_file = self.file_name(output_file + "cache")
                    for p in Path(cache_file).parent.glob(Path(cache_file).name + ".*"):
                        p.unlink()
                    frame = single_snp(
                        test_snps,
                        pheno,
                        cache_file=cache_file,
                        covar=covar,
                        mixing=0,
                        output_file_name=output_file,
                        count_A1=False,
                        leave_out_one_chrom=leave_out_one_chrom,
                        force_full_rank=force_full_rank,
                        force_low_rank=force_low_rank,
                    )

                    self.compare_files(frame, ref_file)
                    frame = single_snp(
                        test_snps,
                        pheno,
                        cache_file=cache_file,
                        covar=covar,
                        mixing=0,
                        output_file_name=output_file,
                        count_A1=False,
                        leave_out_one_chrom=leave_out_one_chrom,
                        force_full_rank=force_full_rank,
                        force_low_rank=force_low_rank,
                    )
                    self.compare_files(frame, ref_file)

    def test_two_looc(self):
        logging.info("TestSingleSnpLeaveOutOneChrom test_two_looc")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("two_looc")
        frame = single_snp(
            test_snps[:, ::10],
            pheno,
            covar=covar,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "two_looc")

    def test_interact_looc(self):
        logging.info("TestSingleSnpLeaveOutOneChrom test_interact_looc")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn

        output_file = self.file_name("interact_looc")
        frame = single_snp(
            test_snps,
            pheno,
            covar=covar,
            mixing=0,
            interact_with_snp=0,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "interact_looc")

    def test_covar_by_chrom(self):
        logging.info("TestSingleSnpLeaveOutOneChrom test_covar_by_chrom")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = Pheno(self.cov_fn).read()
        covar = SnpData(iid=covar.iid, sid=["pheno-1"], val=covar.val)
        covar_by_chrom = {chrom: self.cov_fn for chrom in range(1, 6)}
        output_file = self.file_name("covar_by_chrom")
        frame = single_snp(
            test_snps,
            pheno,
            covar=covar,
            mixing=0,
            covar_by_chrom=covar_by_chrom,
            output_file_name=output_file,
            count_A1=False,
        )

        self.compare_files(frame, "covar_by_chrom")

    def test_covar_by_chrom_mixing(self):
        logging.info("TestSingleSnpLeaveOutOneChrom test_covar_by_chrom_mixing")
        test_snps = Bed(self.bedbase, count_A1=False)
        pheno = self.phen_fn
        covar = self.cov_fn
        covar = Pheno(self.cov_fn).read()
        covar = SnpData(iid=covar.iid, sid=["pheno-1"], val=covar.val)
        covar_by_chrom = {chrom: self.cov_fn for chrom in range(1, 6)}
        output_file = self.file_name("covar_by_chrom_mixing")
        frame = single_snp(
            test_snps,
            pheno,
            covar=covar,
            covar_by_chrom=covar_by_chrom,
            output_file_name=output_file,
            count_A1=False,
        )
        self.compare_files(frame, "covar_by_chrom_mixing")

    def compare_files(self, frame, ref_base):
        reffile = TestFeatureSelection.reference_file("single_snp/" + ref_base + ".txt")

        # sid_list,pvalue_list = frame['SNP'].values,frame['Pvalue'].values

        # sid_to_pvalue = {}
        # for index, sid in enumerate(sid_list):
        #    sid_to_pvalue[sid] = pvalue_list[index]

        reference = pd.read_csv(reffile, delimiter="\s", comment=None, engine="python")
        self.compare_df(frame, reference, reffile)

    def compare_df(self, frame, reference, name):
        assert len(frame) == len(
            reference
        ), "# of pairs differs from file '{0}'".format(name)
        if "Pheno" not in frame.columns or "Pheno" not in reference.columns:
            frame.set_index("SNP", inplace=True)
            reference.set_index("SNP", inplace=True)
        else:
            frame.set_index(["Pheno", "SNP"], inplace=True)
            reference.set_index(["Pheno", "SNP"], inplace=True)

        diff = frame.PValue - reference.PValue
        bad = diff[np.abs(diff) > 1e-5]
        if len(bad) > 0:
            raise Exception(
                "snps differ too much from file '{0}' at these snps {1}".format(
                    name, bad
                )
            )

    def compare_files_effect_size(self, frame, ref_base):
        reffile = TestFeatureSelection.reference_file("single_snp/" + ref_base + ".txt")
        reference = pd.read_csv(reffile, delimiter="\s", comment=None, engine="python")
        self.compare_df_effect_size(frame, reference, reffile)

    def compare_df_effect_size(self, frame, reference, name):
        assert len(frame) == len(
            reference
        ), "# of pairs differs from file '{0}'".format(name)
        if "Pheno" not in frame.columns or "Pheno" not in reference.columns:
            frame.set_index("SNP", inplace=True)
            reference.set_index("SNP", inplace=True)
        else:
            frame.set_index(["Pheno", "SNP"], inplace=True)
            reference.set_index(["Pheno", "SNP"], inplace=True)

        diff = frame.EffectSize - reference.EffectSize
        bad = diff[np.abs(diff) > 1e-5]
        if len(bad) > 0:
            raise Exception(
                "EffectSize of snps differ too much from file '{0}' at these snps {1}".format(
                    name, bad
                )
            )


def getTestSuite():
    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestSingleSnp)
    suite2 = unittest.TestLoader().loadTestsFromTestCase(TestSingleSnpLeaveOutOneChrom)
    return unittest.TestSuite([suite1, suite2])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc, LocalInParts

    # this import is needed for the runner
    from fastlmm.association.tests.test_single_snp import (
        TestSingleSnp,
        TestSingleSnpLeaveOutOneChrom,
    )

    suites = unittest.TestSuite([getTestSuite()])

    if True:  # Standard test run
        r = unittest.TextTestRunner(failfast=False)
        ret = r.run(suites)
        assert ret.wasSuccessful()
    else:  # Cluster test run
        logging.basicConfig(level=logging.INFO)

        from pysnptools.util.mapreduce1.distributabletest import DistributableTest

        # runner = Local()
        runner = LocalMultiProc(taskcount=2, mkl_num_threads=5, just_one_process=False)
        # runner = LocalInParts(0,2,mkl_num_threads=1) # For debugging the cluster runs
        distributable_test = DistributableTest(suites, "temp_test")
        print(runner.run(distributable_test))

    logging.info("done with testing")



import numpy as np
import scipy.linalg as LA
import scipy.optimize as opt
import scipy.stats as ST
import scipy.special as SS
from fastlmm.util.mingrid import *
from fastlmm.util.util import *
import time
import warnings
import logging


class LMM(object):
    """
    linear mixed model with up to two kernels
    N(y | X*beta ; sigma2(h2*((1-a2)*K0 + a2*K1) + (1-h2)*I),
    where
    K0 = G0*G0^T
    K1 = G1*G1^T
    """

    __slots__ = [
        "G",
        "G0",
        "G1",
        "y",
        "X",
        "K0",
        "K1",
        "K",
        "U",
        "S",
        "UX",
        "Uy",
        "UUX",
        "UW",
        "UUW",
        "UUy",
        "pos0",
        "pos1",
        "a2",
        "exclude_idx",
        "forcefullrank",
        "numcalls",
        "Xstar",
        "Kstar",
        "Kstar_star",
        "UKstar",
        "UUKstar",
        "Gstar",
        "K0star",
        "K1star",
        "K0star_star",
        "K1star_star",
    ]

    def __init__(self, forcefullrank=False):
        """
        Input:
        forcefullrank   : if True, then the code always computes K and runs cubically
                            (False)
        """
        self.X = None
        self.y = None
        self.G = None
        self.G0 = None
        self.G1 = None
        self.K = None
        self.K0 = None
        self.K1 = None
        self.U = None
        self.S = None
        self.Uy = None
        self.UUy = None
        self.UX = None
        self.UUX = None
        self.UW = None
        self.UUW = None
        self.pos0 = None
        self.pos1 = None
        self.a2 = None
        self.exclude_idx = []
        self.forcefullrank = forcefullrank
        self.numcalls = 0
        self.Xstar = None
        self.Kstar = None
        self.Kstar_star = None
        self.UKstar = None
        self.UUKstar = None
        self.Gstar = None

    def setX(self, X):
        """
        set the fixed effects X (covariates).
        The Kernel has to be set in advance by first calling setG() or setK().
        --------------------------------------------------------------------------
        Input:
        X       : [N*D] 2-dimensional array of covariates
        --------------------------------------------------------------------------
        """
        self.X = X
        self.UX = self.U.T.dot(X)
        k = self.S.shape[0]
        N = self.X.shape[0]
        if k < N:
            self.UUX = X - self.U.dot(self.UX)

    def setX2(self, X):
        """
        a version of setX that doesn't assume that Eigenvalue decomposition has been done.
        """
        self.X = X
        N = self.X.shape[0]

    def sety(self, y):
        """
        set the phenotype y.
        The Kernel has to be set in advance by first calling setG() or setK().
        --------------------------------------------------------------------------
        Input:
        y       : [N] 1-dimensional array of phenotype values
        --------------------------------------------------------------------------
        """
        assert y.ndim == 1, "y should be 1-dimensional"
        self.y = y
        self.Uy = self.U.T.dot(y)
        k = self.S.shape[0]
        N = self.y.shape[0]
        if k < N:
            self.UUy = y - self.U.dot(self.Uy)

    def sety2(self, y):
        """
        a version of sety that doesn't assume that Eigenvalue decomposition has been done.
        """
        assert y.ndim == 1, "y should be 1-dimensional"
        self.y = y
        N = self.y.shape[0]

    def setG(self, G0=None, G1=None, a2=0.0, K0=None, K1=None):
        """
        set the Kernel (1-a2)*K0 and a2*K1 from G0 and G1.
        This has to be done before setting the data setX() and setY().

        If k0+k1>>N and similar kernels are used repeatedly, it is beneficial to precompute
        the kernel and pass it as an argument.
        ----------------------------------------------------------------------------
        Input:
        G0              : [N*k0] array of random effects
        G1              : [N*k1] array of random effects (optional)
        a2              : mixture weight between K0=G0*G0^T and K1=G1*G1^T

        K0              : [N*N] array, random effects covariance (positive semi-definite)
        K1              : [N*N] array, random effects covariance (positive semi-definite)(optional)
        -----------------------------------------------------------------------------
        """
        self.G0 = G0
        self.G1 = G1
        if a2 < 0.0:
            a2 = 0.0
        if a2 > 1.0:
            a2 = 1.0

        if G1 is None and G0 is not None:
            self.G = G0
        elif G0 is not None and G1 is not None:
            # build the weighted concatenation of G0 and G1 = varianceComponent
            if a2 == 0.0:
                logging.info("a2=0.0, only using G0")
                self.G = G0
            elif a2 == 1.0:
                self.G = G1
                logging.info("a2=1.0, only using G1")
            else:
                self.G = np.concatenate((np.sqrt(1.0 - a2) * G0, np.sqrt(a2) * G1), 1)

        else:
            self.G = None

        if self.G is not None:
            N = self.G.shape[0]
            k = self.G.shape[1]
        else:
            N = K0.shape[0]
            k = N
        if k > 0:
            if (not self.forcefullrank) and (k < N):
                # it is faster using the eigen decomposition of G.T*G but this is more accurate
                try:
                    [U, S, V] = LA.svd(self.G, full_matrices=False)  #!!!use big_svd?
                    if np.any(S < -0.1):
                        logging.warning("kernel contains a negative Eigenvalue")
                    self.U = U
                    self.S = S * S

                except LA.LinAlgError:  # revert to Eigenvalue decomposition
                    logging.warning(
                        "Got SVD exception, trying eigenvalue decomposition of square of G. Note that this is a little bit less accurate"
                    )
                    [S_, V_] = LA.eigh(self.G.T.dot(self.G))
                    if np.any(S_ < -0.1):
                        logging.warning("kernel contains a negative Eigenvalue")
                    S_nonz = S_ > 0
                    self.S = S_[S_nonz]
                    self.S *= N / self.S.sum()
                    self.U = self.G.dot(V_[:, S_nonz] / np.sqrt(self.S))
            else:
                if K0 is None:
                    K0 = self.G0.dot(self.G0.T)
                self.K0 = K0
                if (self.G1 is not None) and (K1 is None):
                    K1 = self.G1.dot(self.G1.T)
                self.setK(K0=K0, K1=K1, a2=a2)
                # K=self.G.dot(self.G.T)
                # self.setK(K)
            self.a2 = a2
            pass
        else:  # rank of kernel = 0 (linear regression case)
            self.S = np.zeros((0))
            self.U = np.zeros_like(self.G)

    def setK(self, K0, K1=None, a2=0.0):
        """
        set the Kernel (1-a2)*K0 and a2*K1.
        This has to be done before setting the data setX() and setY().
        --------------------------------------------------------------------------
        Input:
        K0 : [N*N] array, random effects covariance (positive semi-definite)
        K1 : [N*N] array, random effects covariance (positive semi-definite)(optional)
        a2 : mixture weight between K0 and K1
        --------------------------------------------------------------------------
        """
        self.K0 = K0
        self.K1 = K1
        logging.debug("About to mix K0 and K1")
        if K1 is None:
            self.K = K0
        else:
            self.K = (1.0 - a2) * K0 + a2 * K1
        logging.debug("About to eigh")
        [S, U] = LA.eigh(self.K)
        logging.debug("Done with to eigh")
        if np.any(S < -0.1):
            logging.warning("kernel contains a negative Eigenvalue")

        self.U = U
        self.S = S  # *(S.shape[0]/S.sum())
        self.a2 = a2

    def setK2(self, K0, K1=None, a2=0.0):
        """
        a version of setK that doesn't do Eigenvalue decomposition.
        """
        self.K0 = K0
        self.K1 = K1
        logging.debug("About to mix K0 and K1")
        if K1 is None:
            self.K = K0
        else:
            self.K = (1.0 - a2) * K0 + a2 * K1
        self.a2 = a2

    def set_exclude_idx(self, idx):
        """

        --------------------------------------------------------------------------
        Input:
        idx  : [k_up: number of SNPs to be removed] holds the indices of SNPs to be removed
        --------------------------------------------------------------------------
        """

        self.exclude_idx = idx

    def innerLoopTwoKernel(
        self, a2=0.5, nGridH2=10, minH2=0.0, maxH2=0.99999, **kwargs
    ):
        """
        For a given weight a2, finds the optimal h2 and returns the negative log-likelihood
        --------------------------------------------------------------------------
        Input:
        a2      : mixture weight between K0 and K1
        nGridH2 : number of h2-grid points to evaluate the negative log-likelihood at
        minH2   : minimum value for h2 optimization
        maxH2   : maximum value for h2 optimization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal h2
        --------------------------------------------------------------------------
        """

        if self.K0 is not None:
            self.setK(K0=self.K0, K1=self.K1, a2=a2)
        else:
            self.setG(G0=self.G0, G1=self.G1, a2=a2)
        self.setX(self.X)
        self.sety(self.y)
        return self.findH2(nGridH2=nGridH2, minH2=minH2, maxH2=maxH2, **kwargs)

    def findA2(
        self,
        nGridA2=10,
        minA2=0.0,
        maxA2=1.0,
        nGridH2=10,
        minH2=0.0,
        maxH2=0.99999,
        verbose=False,
        **kwargs
    ):
        """
        Find the optimal a2 and h2, such that K=(1.0-a2)*K0+a2*K1. Performs a double loop optimization (could be expensive for large grid-sizes)
        (default maxA2 value is set to 1 as loss of positive definiteness of the final model covariance only depends on h2, not a2)
        --------------------------------------------------------------------------
        Input:
        nGridA2 : number of a2-grid points to evaluate the negative log-likelihood at
        minA2   : minimum value for a2 optimization
        maxA2   : maximum value for a2 optimization
        nGridH2 : number of h2-grid points to evaluate the negative log-likelihood at
        minH2   : minimum value for h2 optimization
        maxH2   : maximum value for h2 optimization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal h2 and a2
        --------------------------------------------------------------------------
        """
        self.numcalls = 0
        resmin = [None]

        def f(x, resmin=resmin, nGridH2=nGridH2, minH2=minH2, maxH2=maxH2, **kwargs):
            self.numcalls += 1
            t0 = time.time()
            res = self.innerLoopTwoKernel(
                a2=x, nGridH2=nGridH2, minH2=minH2, maxH2=maxH2, **kwargs
            )
            if (resmin[0] is None) or (res["nLL"] < resmin[0]["nLL"]):
                resmin[0] = res
            t1 = time.time()
            logging.info(
                "x={0}. one objective function call took {1} seconds elapsed ".format(
                    x, t1 - t0
                )
            )
            # import pdb; pdb.set_trace()
            return res["nLL"]

        if verbose:
            logging.info("finda2")
        min = minimize1D(f=f, nGrid=nGridA2, minval=minA2, maxval=maxA2, verbose=False)
        # print "numcalls to innerLoopTwoKernel= " + str(self.numcalls)
        return resmin[0]

    def findH2(self, nGridH2=10, minH2=0.0, maxH2=0.99999, REML=True, **kwargs):
        """
        Find the optimal h2 for a given K. Note that this is the single kernel case. So there is no a2.
        (default maxH2 value is set to a value smaller than 1 to avoid loss of positive definiteness of the final model covariance)
        --------------------------------------------------------------------------
        Input:
        nGridH2 : number of h2-grid points to evaluate the negative log-likelihood at
        minH2   : minimum value for h2 optimization
        maxH2   : maximum value for h2 optimization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal h2
        --------------------------------------------------------------------------
        """
        # f = lambda x : (self.nLLeval(h2=x,**kwargs)['nLL'])
        resmin = [None]

        def f(x, resmin=resmin, **kwargs):
            res = self.nLLeval(h2=x, REML=REML, **kwargs)
            if (resmin[0] is None) or (res["nLL"] < resmin[0]["nLL"]):
                resmin[0] = res
            logging.debug("search\t{0}\t{1}".format(x, res["nLL"]))
            return res["nLL"]

        min = minimize1D(f=f, nGrid=nGridH2, minval=minH2, maxval=maxH2)
        return resmin[0]

    def find_log_delta(
        self,
        sid_count,
        min_log_delta=-5,
        max_log_delta=10,
        nGrid=10,
        REML=True,
        **kwargs
    ):
        """
        #Need comments
        """
        # f = lambda x : (self.nLLeval(h2=x,**kwargs)['nLL'])
        resmin = [None]

        def f(x, resmin=resmin, **kwargs):
            h2 = 1.0 / (
                np.exp(x) * sid_count + 1
            )  # We convert from external log_delta to h2 and then back again so that this code is most similar to findH2

            res = self.nLLeval(h2=h2, REML=REML, **kwargs)
            if (resmin[0] is None) or (res["nLL"] < resmin[0]["nLL"]):
                resmin[0] = res
            # logging.info("search\t{0}\t{1}".format(x,res['nLL']))
            return res["nLL"]

        min = minimize1D(f=f, nGrid=nGrid, minval=min_log_delta, maxval=max_log_delta)
        res = resmin[0]
        internal_delta = 1.0 / res["h2"] - 1.0
        ln_external_delta = np.log(internal_delta / sid_count)
        res["log_delta"] = ln_external_delta
        return res

    def nLLeval(
        self,
        h2=0.0,
        REML=True,
        logdelta=None,
        delta=None,
        dof=None,
        scale=1.0,
        penalty=0.0,
    ):
        """
        evaluate -ln( N( U^T*y | U^T*X*beta , h2*S + (1-h2)*I ) ),
        where ((1-a2)*K0 + a2*K1) = USU^T
        --------------------------------------------------------------------------
        Input:
        h2      : mixture weight between K and Identity (environmental noise)
        REML    : boolean
                  if True   : compute REML
                  if False  : compute ML
        dof     : Degrees of freedom of the Multivariate student-t
                        (default None uses multivariate Normal likelihood)
        logdelta: log(delta) allows to optionally parameterize in delta space
        delta   : delta     allows to optionally parameterize in delta space
        scale   : Scale parameter the multiplies the Covariance matrix (default 1.0)
        --------------------------------------------------------------------------
        Output dictionary:
        'nLL'       : negative log-likelihood
        'sigma2'    : the model variance sigma^2
        'beta'      : [D*1] array of fixed effects weights beta
        'h2'        : mixture weight between Covariance and noise
        'REML'      : True: REML was computed, False: ML was computed
        'a2'        : mixture weight between K0 and K1
        'dof'       : Degrees of freedom of the Multivariate student-t
                        (default None uses multivariate Normal likelihood)
        'scale'     : Scale parameter that multiplies the Covariance matrix (default 1.0)
        --------------------------------------------------------------------------
        """
        if (h2 < 0.0) or (h2 > 1.0):
            return {"nLL": 3e20, "h2": h2, "REML": REML, "scale": scale}
        k = self.S.shape[0]
        N = self.y.shape[0]
        D = self.UX.shape[1]

        # if REML == True:
        #    # this needs to be fixed, please see test_gwas.py for details
        #    raise NotImplementedError("REML in lmm object not supported, please use lmm_cov.py instead")

        if logdelta is not None:
            delta = np.exp(logdelta)

        if delta is not None:
            Sd = (self.S + delta) * scale
        else:
            Sd = (h2 * self.S + (1.0 - h2)) * scale

        UXS = self.UX / np.lib.stride_tricks.as_strided(
            Sd, (Sd.size, self.UX.shape[1]), (Sd.itemsize, 0)
        )
        UyS = self.Uy / Sd

        XKX = UXS.T.dot(self.UX)
        XKy = UXS.T.dot(self.Uy)
        yKy = UyS.T.dot(self.Uy)

        logdetK = np.log(Sd).sum()

        if k < N:  # low rank part
            # determine normalization factor
            if delta is not None:
                denom = delta * scale
            else:
                denom = (1.0 - h2) * scale

            XKX += self.UUX.T.dot(self.UUX) / (denom)
            XKy += self.UUX.T.dot(self.UUy) / (denom)
            yKy += self.UUy.T.dot(self.UUy) / (denom)
            logdetK += (N - k) * np.log(denom)

        # proximal contamination (see Supplement Note 2: An Efficient Algorithm for Avoiding Proximal Contamination)
        # available at: http://www.nature.com/nmeth/journal/v9/n6/extref/nmeth.2037-S1.pdf
        # exclude SNPs from the RRM in the likelihood evaluation

        if len(self.exclude_idx) > 0:
            num_exclude = len(self.exclude_idx)

            # consider only excluded SNPs
            G_exclude = self.G[:, self.exclude_idx]

            self.UW = self.U.T.dot(G_exclude)  # needed for proximal contamination
            UWS = self.UW / np.lib.stride_tricks.as_strided(
                Sd, (Sd.size, num_exclude), (Sd.itemsize, 0)
            )
            assert UWS.shape == (k, num_exclude)

            WW = np.eye(num_exclude) - UWS.T.dot(self.UW)
            WX = UWS.T.dot(self.UX)
            Wy = UWS.T.dot(self.Uy)
            assert WW.shape == (num_exclude, num_exclude)
            assert WX.shape == (num_exclude, D)
            assert Wy.shape == (num_exclude,)

            if k < N:  # low rank part
                self.UUW = G_exclude - self.U.dot(self.UW)

                WW += self.UUW.T.dot(self.UUW) / denom
                WX += self.UUW.T.dot(self.UUX) / denom
                Wy += self.UUW.T.dot(self.UUy) / denom

            # TODO: do cholesky, if fails do eigh
            # compute inverse efficiently
            [S_WW, U_WW] = LA.eigh(WW)

            UWX = U_WW.T.dot(WX)
            UWy = U_WW.T.dot(Wy)
            assert UWX.shape == (num_exclude, D)
            assert UWy.shape == (num_exclude,)

            # compute S_WW^{-1} * UWX
            WX = UWX / np.lib.stride_tricks.as_strided(
                S_WW, (S_WW.size, UWX.shape[1]), (S_WW.itemsize, 0)
            )
            # compute S_WW^{-1} * UWy
            Wy = UWy / S_WW
            # determinant update
            logdetK += np.log(S_WW).sum()
            assert WX.shape == (num_exclude, D)
            assert Wy.shape == (num_exclude,)

            # perform updates (instantiations for a and b in Equation (1.5) of Supplement)
            yKy += UWy.T.dot(Wy)
            XKy += UWX.T.dot(Wy)
            XKX += UWX.T.dot(WX)

        #######

        [SxKx, UxKx] = LA.eigh(XKX)
        # optionally regularize the beta weights by penalty
        if penalty > 0.0:
            SxKx += penalty
        i_pos = SxKx > 1e-10
        beta = UxKx[:, i_pos].dot(UxKx[:, i_pos].T.dot(XKy) / SxKx[i_pos])

        r2 = yKy - XKy.dot(beta)

        if dof is None:  # Use the Multivariate Gaussian
            if REML:
                XX = self.X.T.dot(self.X)
                [Sxx, Uxx] = LA.eigh(XX)
                logdetXX = np.log(Sxx).sum()
                logdetXKX = np.log(SxKx).sum()
                sigma2 = r2 / (N - D)
                nLL = 0.5 * (
                    logdetK
                    + logdetXKX
                    - logdetXX
                    + (N - D) * (np.log(2.0 * np.pi * sigma2) + 1)
                )
                variance_beta = None
            else:
                sigma2 = r2 / (N)
                nLL = 0.5 * (logdetK + N * (np.log(2.0 * np.pi * sigma2) + 1))
                if delta is not None:
                    h2 = 1.0 / (delta + 1)
                # This is a faster version of h2 * sigma2 * np.diag(LA.inv(XKX))
                # where h2*sigma2 is sigma2_g
                variance_beta = (
                    h2
                    * sigma2
                    * (UxKx[:, i_pos] / SxKx[i_pos] * UxKx[:, i_pos]).sum(-1)
                )
            result = {
                "nLL": nLL,
                "sigma2": sigma2,
                "beta": beta,
                "variance_beta": variance_beta,
                "h2": h2,
                "REML": REML,
                "a2": self.a2,
                "scale": scale,
            }
        else:  # Use multivariate student-t
            if REML:
                XX = self.X.T.dot(self.X)
                [Sxx, Uxx] = LA.eigh(XX)
                logdetXX = np.log(Sxx).sum()
                logdetXKX = np.log(SxKx).sum()

                nLL = 0.5 * (
                    logdetK
                    + logdetXKX
                    - logdetXX
                    + (dof + (N - D)) * np.log(1.0 + r2 / dof)
                )
                nLL += (
                    0.5 * (N - D) * np.log(dof * np.pi)
                    + SS.gammaln(0.5 * dof)
                    - SS.gammaln(0.5 * (dof + (N - D)))
                )
            else:
                nLL = 0.5 * (logdetK + (dof + N) * np.log(1.0 + r2 / dof))
                nLL += (
                    0.5 * N * np.log(dof * np.pi)
                    + SS.gammaln(0.5 * dof)
                    - SS.gammaln(0.5 * (dof + N))
                )
            result = {
                "nLL": nLL,
                "dof": dof,
                "beta": beta,
                "variance_beta": None,
                "h2": h2,
                "REML": REML,
                "a2": self.a2,
                "scale": scale,
            }
        assert np.all(
            np.isreal(nLL)
        ), "nLL has an imaginary component, possibly due to constant covariates"
        if result["variance_beta"] is None:
            del result["variance_beta"]
        return result

    def getPosteriorWeights(self, beta, h2=0.0, logdelta=None, delta=None, scale=1.0):
        """
        compute posterior mean over the feature weights (effect sizes of SNPs in the kernel, not the SNPs being tested):
        w = G.T (GG.T + delta*I)^(-1) (y - Xbeta)
        --------------------------------------------------------------------------
        Input:
        beta            : weight vector for fixed effects
        h2              : mixture weight between K and Identity (environmental noise)
        logdelta        : log(delta) allows to optionally parameterize in delta space
        delta           : delta     allows to optionally parameterize in delta space
        scale           : Scale parameter the multiplies the Covariance matrix (default 1.0)

        returnVar       : if True, marginal variances are estimated
        returnCovar     : if True, posterior covariance is learnt
        --------------------------------------------------------------------------
        Dictionary with the following fields:
        weights           : [k0+k1] 1-dimensional array of predicted phenotype values
        --------------------------------------------------------------------------
        """
        k = self.S.shape[0]
        N = self.y.shape[0]

        if logdelta is not None:
            delta = np.exp(logdelta)
        if delta is not None:
            Sd = (self.S + delta) * scale
        else:
            Sd = (h2 * self.S + (1.0 - h2)) * scale

        yres = self.y - np.dot(self.X, beta)
        Uyres = np.dot(self.U.T, yres)
        UG = np.dot(self.U.T, self.G)
        weights = np.dot(UG.T, Uyres / Sd)

        if k < N:  # low-rank part
            # determine normalization factor
            if delta is not None:
                denom = delta * scale
            else:
                denom = (1.0 - h2) * scale

            UUG = self.G - np.dot(self.U, UG)
            UUyres = yres - np.dot(self.U, Uyres)
            weights += UUG.T.dot(UUyres) / (denom)

        return weights

    def setTestData(self, Xstar, K0star=None, K1star=None, G0star=None, G1star=None):
        """
        set data for predicting

        --------------------------------------------------------------------------
        Input:
        Xstar           : [M,D] 2-dimensional array of covariates on the test set
        G0star          : [M,k0] array of random effects on the test set
        G1star          : [M,k1] array of random effects on the test set (optional)
        K0star          : [M,N] array, random effects covariance between test and training data (positive semi-definite)
        K1star          : [M,N] array, random effects covariance between test and training data (positive semi-definite)(optional)
        where M is # of test cases, N is the # of training cases
        --------------------------------------------------------------------------
        """

        self.Xstar = Xstar
        if G1star is None:
            self.Gstar = G0star
        else:
            if self.a2 == 0.0:
                logging.info("a2=0.0, only using G0")
                self.Gstar = G0star
            elif self.a2 == 1.0:
                self.Gstar = G1star
                logging.info("a2=1.0, only using G1")
            else:
                self.Gstar = np.concatenate(
                    (np.sqrt(1.0 - self.a2) * G0star, np.sqrt(self.a2) * G1star), 1
                )

        if K0star is not None:
            if K1star is None:
                self.Kstar = K0star
            else:
                self.Kstar = (1.0 - self.a2) * K0star + self.a2 * K1star
        else:
            self.Kstar = np.dot(self.Gstar, self.G.T)

        self.UKstar = np.dot(self.U.T, self.Kstar.T)

        if self.G is not None:
            k = self.G.shape[1]
            N = self.G.shape[0]
            if k < N:
                # see e.g. Equation 3.17 in Supplement of FaST LMM paper
                self.UUKstar = self.Kstar.T - np.dot(self.U, self.UKstar)

    def setTestData2(self, Xstar, K0star=None, K1star=None):
        """
        a version of setTestData that doesn't assume that Eigenvalue decomposition has been done.
        """

        self.Xstar = Xstar
        self.Gstar = None
        if K1star is None:
            self.Kstar = K0star
        else:
            self.Kstar = (1.0 - self.a2) * K0star + self.a2 * K1star

    def predictMean(self, beta, h2=0.0, logdelta=None, delta=None, scale=1.0):
        """
        mean prediction for the linear mixed model on unobserved data:

        ystar = X*beta + Kstar(h2*K + (1-h2)*K)^{-1}(y-X*beta)
        where Kstar is the train vs test kernel
        --------------------------------------------------------------------------
        Input:
        beta            : weight vector for fixed effects
        h2              : mixture weight between K and Identity (environmental noise)
        logdelta        : log(delta) allows to optionally parameterize in delta space
        delta           : delta     allows to optionally parameterize in delta space
        scale           : Scale parameter the multiplies the Covariance matrix (default 1.0)


        If SNPs are excluded, nLLeval must be called before to re-calculate self.UW,self.UUW
        --------------------------------------------------------------------------
        Output:
        ystar           : [M] 1-dimensional array of predicted phenotype values
        --------------------------------------------------------------------------
        """

        M = self.Xstar.shape[0]

        if (h2 < 0.0) or (h2 >= 1.0):
            return np.nan * np.ones(M)

        k = self.S.shape[0]
        N = self.y.shape[0]
        # D=self.UX.shape[1]

        if logdelta is not None:
            delta = np.exp(logdelta)

        # delta = (1-h2) / h2
        if delta is not None:
            Sd = (self.S + delta) * scale
        else:
            assert False, "not implemented (UKstar needs to be scaled by h2)"
            Sd = (h2 * self.S + (1.0 - h2)) * scale

        if len(self.exclude_idx) > 0:
            # cut out
            num_exclude = len(self.exclude_idx)
            # consider only excluded SNPs
            Gstar_exclude = self.Gstar[:, self.exclude_idx]
            # G_exclude = self.G[:,self.exclude_idx]
            UKstar = self.UKstar - np.dot(self.UW, Gstar_exclude.T)
            if k < N:
                UUKstar = self.UUKstar - np.dot(self.UUW, Gstar_exclude.T)
        else:
            UKstar = self.UKstar
            UUKstar = self.UUKstar

        yfixed = np.dot(self.Xstar, beta)
        yres = self.y - np.dot(self.X, beta)
        Uyres = self.Uy - np.dot(self.UX, beta)
        Sdi = 1.0 / Sd
        yrandom = np.dot(Sdi * UKstar.T, Uyres)

        if k < N:  # low-rank part
            # determine normalization factor
            if delta is not None:
                denom = delta * scale
            else:
                denom = (1.0 - h2) * scale
            UUyres = yres - np.dot(self.U, Uyres)
            yrandom += np.dot(UUKstar.T, UUyres) / denom

        # proximal contamination (see Supplement Note 2: An Efficient Algorithm for Avoiding Proximal Contamination)
        # available at: http://www.nature.com/nmeth/journal/v9/n6/extref/nmeth.2037-S1.pdf
        # exclude SNPs from the RRM in the likelihood evaluation
        if len(self.exclude_idx) > 0:
            UWS = self.UW / np.lib.stride_tricks.as_strided(
                Sd, (Sd.size, num_exclude), (Sd.itemsize, 0)
            )
            assert UWS.shape == (k, num_exclude)
            WW = np.eye(num_exclude) - UWS.T.dot(self.UW)
            WKstar = UWS.T.dot(UKstar)
            Wyres = UWS.T.dot(Uyres)
            assert WW.shape == (num_exclude, num_exclude)
            assert WKstar.shape == (num_exclude, M)
            assert Wyres.shape == (num_exclude,)

            if k < N:  # low rank part
                WW += self.UUW.T.dot(self.UUW) / denom
                WKstar += self.UUW.T.dot(UUKstar) / denom
                Wyres += self.UUW.T.dot(UUyres) / denom

            # TODO: do cholesky, if fails do eigh
            # compute inverse efficiently
            [S_WW, U_WW] = LA.eigh(WW)

            UWKstar = U_WW.T.dot(WKstar)
            UWyres = U_WW.T.dot(Wyres)
            assert UWKstar.shape == (num_exclude, M)
            assert UWyres.shape == (num_exclude,)

            # compute S_WW^{-1} * UWX
            WKstar = UWKstar / np.lib.stride_tricks.as_strided(
                S_WW, (S_WW.size, UWKstar.shape[1]), (S_WW.itemsize, 0)
            )
            # compute S_WW^{-1} * UWy
            Wyres = UWyres / S_WW
            assert WKstar.shape == (num_exclude, M)
            assert Wyres.shape == (num_exclude,)

            # perform updates (instantiations for a and b in Equation (1.5) of Supplement)
            yrandom += UWKstar.T.dot(Wyres)

        ystar = yfixed + yrandom
        return ystar

    def predict_mean_and_variance(lmm, beta, sigma2, h2, Kstar_star):
        assert 0 <= h2 <= 1, "By definition, h2 must be between 0 and 1 (inclusive)"
        varg = h2 * sigma2
        vare = (1.0 - h2) * sigma2
        if lmm.G is not None:
            K = np.dot(
                lmm.G, lmm.G.T
            )  #!!!later this is very inefficient in memory and computation
        else:
            K = np.dot(
                np.dot(lmm.U, np.eye(len(lmm.U)) * lmm.S), lmm.U.T
            )  # Re-compose the Eigen value decomposition #!!!later do this more efficiently
        V = varg * K + vare * np.eye(len(K))
        Vinv = LA.inv(V)

        a = np.dot(varg * lmm.Kstar, Vinv)

        y_star = np.dot(lmm.Xstar, beta) + np.dot(
            a, lmm.y - np.dot(lmm.X, beta)
        )  #!!!later shouldn't the 2nd dot be precomputed?
        y_star = y_star.reshape(-1, 1)  # Make 2-d

        var_star = (
            varg * Kstar_star
            + vare * np.eye(len(Kstar_star))
            - np.dot(a, (varg * lmm.Kstar.T))
        )
        return y_star, var_star

    def nLL(lmm, beta, sigma2, h2, y_actual):
        from scipy.stats import multivariate_normal

        y_star, var_star = predict_mean_and_variance(
            lmm, beta, sigma2, h2, lmm.Kstar_star
        )
        var = multivariate_normal(mean=y_star.reshape(-1), cov=var_star)
        return -np.log(var.pdf(y_actual.reshape(-1)))

    def predictVariance(
        self, h2=0.0, logdelta=None, delta=None, sigma2=1.0, Kstar_star=None
    ):
        """
        variance prediction for the linear mixed model on unobserved data:

        Var_star = sigma2 * (K(X*,X*) + delta*I - Kstar (K + delta*I)^{-1} Kstar )
        --------------------------------------------------------------------------
        Input:
        h2              : mixture weight between K and Identity (environmental noise)
        logdelta        : log(delta) allows to optionally parameterize in delta space
        delta           : delta     allows to optionally parameterize in delta space
        sigma2          : sigma2 parameter the multiplies the Covariance matrix (default 1.0)
        K_star_star     : Kernel on test examples

        If SNPs are excluded, nLLeval must be called before to re-calculate self.UW,self.UUW
        --------------------------------------------------------------------------
        Output:
        Cov_star           : [M,M] 2-dimensional array covariance matrix
        --------------------------------------------------------------------------
        """

        # TODO: proximal contamination
        # TODO: REML?

        if (h2 < 0.0) or (h2 >= 1.0):
            return np.nan * np.ones(M)

        k = self.S.shape[0]
        N = self.y.shape[0]
        # D = self.UX.shape[1]

        # print "k, N, D", k, N, D

        if logdelta is not None:
            delta = np.exp(logdelta)

        if delta is not None:
            # Sd = (self.S+delta)*sigma2
            Sd = self.S + delta
        else:
            # Sd = (h2*self.S + (1.0-h2))*sigma2
            Sd = h2 * self.S + (1.0 - h2)
            assert False, "h2 code path not test. Please use delta or logdelta"
            # delta = 1.0/h2-1.0 #right?

        Sdi = 1.0 / Sd

        # part 1 from c-code
        # TODO: handle h2 parameterization
        # TODO: make more efficient (add_diag)

        if Kstar_star is None:
            N_test = self.Gstar.shape[0]
            Kstar_star = np.dot(self.Gstar, self.Gstar.T)
        else:
            Kstar_star = Kstar_star.copy()

        N_test = Kstar_star.shape[0]
        assert N_test == Kstar_star.shape[1]

        part1 = Kstar_star
        part1 += np.eye(N_test) * delta
        part1 *= sigma2

        # print "part1", part1[0,0]
        # print "delta", delta, "sigma2", sigma2

        # part 2 from c-code
        # (U1^T a)^T (S_1 + delta*I)^{-1} (U1^T a)
        SUKstarTUkStar = np.dot(Sdi * self.UKstar.T, self.UKstar)

        # UXS = self.UKstar / np.lib.stride_tricks.as_strided(Sd, (Sd.size,self.UKstar.shape[1]), (Sd.itemsize,0))
        # np.testing.assert_array_almost_equal(SUKstarTUkStar, np.dot(UXS.T, self.UKstar), decimal=4)

        SUKstarTUkStar *= sigma2

        # print "UKstar[0,0]", self.UKstar[0,0]
        # print "UKstarS[0,0]", UXS[0,0]
        # print "SUK", SUKstarTUkStar[0,0]

        # part 3&4 from c-code
        if k < N:  # low-rank part
            # determine normalization factor
            if delta is not None:
                denom = delta * sigma2
            else:
                denom = (1.0 - h2) * sigma2

            # see Equation 3.17 in Supplement of FaST LMM paper:
            # 1 / delta * (((I_n - U1U1^T)a)^T (I_n - U1U1^T)a), a=K(XS,X)
            SUKstarTUkStar += np.dot(self.UUKstar.T, self.UUKstar) / denom

        # see Carl Rasmussen's book on GPs, Equation 2.24
        # or Equation 5 in Lasso-LMM paper
        Var_star = part1 - SUKstarTUkStar

        return Var_star

    def nLLeval_test(
        self,
        y_test,
        beta,
        h2=0.0,
        logdelta=None,
        delta=None,
        sigma2=1.0,
        Kstar_star=None,
        robust=False,
    ):
        """
        compute out-of-sample log-likelihood

        robust: boolean
                indicates if eigenvalues will be truncated at 1E-9 or 1E-4. The former (default) one was used in FastLMMC,
                but may lead to numerically unstable solutions.
        """
        assert y_test.ndim == 1, "y_test should have 1 dimension"
        mu = self.predictMean(beta, h2=h2, logdelta=logdelta, delta=delta)
        res = y_test - mu

        sigma = self.predictVariance(
            h2=h2, logdelta=logdelta, delta=delta, sigma2=sigma2, Kstar_star=Kstar_star
        )

        # TODO: benchmark, record speed difference
        """
        # efficient computation of: (y - mu)^T sigma2^{-1} (y - mu)
        # Solve the linear system x = (L L^T)^-1 res

        try:
            L = np.linalg.cho_factor(sigma)
            res_sig = np.linalg.cho_solve(L, res)
            logdetK = np.linalg.slogdet(sigma)[1]

        except Exception, detail:
            print "Cholesky failed, using eigen-value decomposition!"
        """

        [S_, U_] = LA.eigh(sigma)

        if robust:
            S_nonz = S_ > 1e-4
        else:
            S_nonz = S_ > 1e-9
        assert sum(S_nonz) > 0, "Some eigenvalues should be nonzero"
        S = S_[S_nonz]
        U = U_[:, S_nonz]
        Sdi = 1 / S

        res_sig = res.T.dot(Sdi * U).dot(U.T)
        logdetK = np.log(S).sum()

        # some sanity checks
        if False:
            res_sig3 = np.linalg.pinv(sigma).dot(res)
            np.testing.assert_array_almost_equal(res_sig, res_sig3, decimal=2)

        # see Carl Rasmussen's book on GPs, equation 5.10, or
        term1 = -0.5 * logdetK
        term2 = -0.5 * np.dot(
            res_sig.reshape(-1).T, res.reshape(-1)
        )  # Change the inputs to the functions so that these are vectors, not 1xn,nx1
        term3 = -0.5 * len(res) * np.log(2 * np.pi)

        if term2 < -10000:
            logging.warning(
                "looks like nLLeval_test is running into numerical difficulties"
            )

            SC = S.copy()
            SC.sort()

            logging.warning(
                ["delta:", delta, "log det", logdetK, "term 2", term2, "term 3:", term3]
            )
            logging.warning(
                [
                    "largest eigv:",
                    SC[-1],
                    "second largest eigv:",
                    SC[-2],
                    "smallest eigv:",
                    SC[0],
                ]
            )
            logging.warning(
                [
                    "ratio 1large/2large:",
                    SC[-1] / SC[-2],
                    "ratio lrg/small:",
                    SC[-1] / SC[0],
                ]
            )

        neg_log_likelihood = -(term1 + term2 + term3)

        return neg_log_likelihood



import numpy as np
import logging
import unittest
import os.path
import time
import sys
import doctest
import pandas as pd
import fastlmm.util.util as ut
from fastlmm.association.heritability_spatial_correction import (
    heritability_spatial_correction,
)
from pysnptools.util.mapreduce1.runner import Local, LocalMultiProc
from pysnptools.snpreader import Dat, Bed, Pheno, SnpData
from fastlmm.feature_selection.test import TestFeatureSelection


tolerance = 1e-4


class TestHeritabilitySpatialCorrection(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        from pysnptools.util import create_directory_if_necessary

        create_directory_if_necessary(self.tempout_dir, isfile=False)
        self.pythonpath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "..")
        )
        self.snpreader_whole = Bed(
            self.pythonpath + "/tests/datasets/synth/all.bed", count_A1=False
        )
        self.pheno_whole = Pheno(
            self.pythonpath + "/tests/datasets/synth/pheno_10_causals.txt"
        )

    tempout_dir = "tempout/heritability_spatial_correction"

    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name)
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    def test_one(self):
        """
        Lock in results on arbitrary data -- because meaningful runs take too long to run.
        """
        fn = "one.txt"
        logging.info(fn)
        tmpOutfile = self.file_name(fn)

        half = self.pheno_whole.read().val
        pheno = SnpData(
            iid=self.pheno_whole.iid, sid=["pheno0", "pheno1"], val=np.c_[half, half]
        )

        spatial_coor = [[i, -i] for i in range(self.snpreader_whole.iid_count)]
        alpha_list = alpha_list_big = [
            int(v) for v in np.logspace(2, np.log10(4000), 2)
        ]
        dataframe = heritability_spatial_correction(
            self.snpreader_whole,
            spatial_coor,
            self.snpreader_whole.iid,
            alpha_list,
            2,
            pheno,
            jackknife_count=2,
            permute_plus_count=1,
            permute_times_count=1,
            just_testing=True,
        )

        dataframe.to_csv(tmpOutfile, sep="\t", index=False)
        referenceOutfile = TestFeatureSelection.reference_file(
            "heritability_spatial_correction/" + fn
        )
        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out,
            "msg='{0}', ref='{1}', tmp='{2}'".format(msg, referenceOutfile, tmpOutfile),
        )

    def test_two(self):
        """
        Lock in results on arbitrary data -- because meaningful runs take too long to run.
        """
        fn = "two.txt"
        logging.info(fn)
        tmpOutfile = self.file_name(fn)

        snpreader = self.snpreader_whole[:10, :]

        spatial_coor = [[i, -i] for i in range(snpreader.iid_count)]
        alpha_list = alpha_list_big = [
            int(v) for v in np.logspace(2, np.log10(4000), 2)
        ]
        dataframe = heritability_spatial_correction(
            snpreader,
            spatial_coor,
            snpreader.iid,
            alpha_list,
            2,
            self.pheno_whole,
            jackknife_count=2,
            permute_plus_count=1,
            permute_times_count=1,
            just_testing=False,
        )

        dataframe.to_csv(tmpOutfile, sep="\t", index=False)
        referenceOutfile = TestFeatureSelection.reference_file(
            "heritability_spatial_correction/" + fn
        )
        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out,
            "msg='{0}', ref='{1}', tmp='{2}'".format(msg, referenceOutfile, tmpOutfile),
        )

    def test_three(self):
        """
        Lock in results on arbitrary data -- because meaningful runs take too long to run.
        """
        fn = "three.txt"
        logging.info(fn)
        tmpOutfile = self.file_name(fn)

        snpreader = self.snpreader_whole[:10, :]

        spatial_coor = [[i, -i] for i in range(snpreader.iid_count)]
        alpha_list = alpha_list_big = [
            int(v) for v in np.logspace(2, np.log10(4000), 2)
        ]
        dataframe = heritability_spatial_correction(
            snpreader,
            spatial_coor,
            snpreader.iid,
            alpha_list,
            2,
            self.pheno_whole,
            jackknife_count=0,
            permute_plus_count=0,
            permute_times_count=0,
            just_testing=False,
        )

        dataframe.to_csv(tmpOutfile, sep="\t", index=False)
        referenceOutfile = TestFeatureSelection.reference_file(
            "heritability_spatial_correction/" + fn
        )
        out, msg = ut.compare_files(tmpOutfile, referenceOutfile, tolerance)
        self.assertTrue(
            out,
            "msg='{0}', ref='{1}', tmp='{2}'".format(msg, referenceOutfile, tmpOutfile),
        )

    def test_doctest(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        result = doctest.testmod(
            sys.modules["fastlmm.association.heritability_spatial_correction"]
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():

    suite1 = unittest.TestLoader().loadTestsFromTestCase(
        TestHeritabilitySpatialCorrection
    )
    return unittest.TestSuite([suite1])


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    # this import is needed for the runner
    # from fastlmm.association.tests.test_heritability_spatical_correction import TestHeritabilitySpatialCorrection
    suites = unittest.TestSuite([getTestSuite()])

    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()
    logging.info("done with testing")



import warnings
import numpy as np
import sys
import logging
import time
from fastlmm.external.pca import PCA
from pysnptools.snpreader import Bed
from unittest.mock import patch


def compute_auto_pcs(
    snpreader, cutoff=0.1, k_values=np.arange(11), output_file_name=None, count_A1=None
):
    """
    DEPRECATED: Runs only on Intel, not AMD. We can't test this function because `compute_auto_pcs` uses the 'fastlmmC' executable which is not available in the test environment.

    Function automatically finds the best principle components (PCs)

    :param snpreader: SNPs for which to find the best PCs
          If you give a string, it should be the base name of a set of PLINK Bed-formatted files.
    :type snpreader: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__ or a string

    :param cutoff: (default: .1) The degree of relatedness to remove before finding the best number of PCs.
        Relatedness is measured with a RRM (realized relationship matrix) so 0 is no relation, .5 is a sibling or parent, and 1 is self or twin.
    :type cutoff: a number between 0 and 1.

    :param k_values: (default: 0 ... 10 [inclusive]) The number of PCs to search.
    :type k_values: list of integers

    :param count_A1: If it needs to read SNP data from a BED-formatted file, tells if it should count the number of A1
         alleles (the PLINK standard) or the number of A2 alleles. False is the current default, but in the future the default will change to True.
    :type count_A1: bool

    :rtype: A phenotype dictionary with property 'iid' listing the iids and property 'vals' containing a nparray of PC values.

    :Example:

    >>> import logging
    >>> from fastlmm.util import compute_auto_pcs
    >>> from fastlmm.util import example_file # Download and return local file name
    >>> logging.basicConfig(level=logging.INFO)
    >>> file_name = example_file("fastlmm/feature_selection/examples/toydata.5chrom.*","*.bed")
    >>> best_pcs = compute_auto_pcs(file_name,count_A1=False)
    >>> print(int(best_pcs['vals'].shape[0]),int(best_pcs['vals'].shape[1]))
    500 0

    """
    warning_text = "This function is deprecated. Runs only on Intel, not AMD. We can't test this function because `compute_auto_pcs` uses the 'fastlmmC' executable which is not available in the test environment."
    warnings.warn(warning_text, category=DeprecationWarning, stacklevel=2)
    print(warning_text, file=sys.stderr)
    with patch.dict("os.environ", {"ARRAY_MODULE": "numpy"}) as _:
        #!!could use regression tests beyond the docttest

        snpreader = _snp_fixup(snpreader, count_A1=count_A1)

        logging.info("reading all_std_snpdata")
        all_std_snpdata = (
            snpreader.read().standardize()
        )  #!!could doing C or F be better?

        # use vertex cut to find just parents
        logging.info("Finding relatedness of all iids")
        from pysnptools.standardizer import Identity

        rrm = all_std_snpdata.kernel(Identity()) / snpreader.sid_count
        import fastlmm.util.VertexCut as vc

        remove_set = set(
            vc.VertexCut().work(rrm, cutoff)
        )  # These are the indexes of the IIDs to remove
        logging.info(
            "removing {0} of {1} iids".format(len(remove_set), snpreader.iid_count)
        )
        keep_list = [x for x in range(all_std_snpdata.iid_count) if x not in remove_set]
        nofam_snpreader = all_std_snpdata[keep_list, :]
        # nofam_snpreader = all_std_snpdata#[1:,:]
        # print "#!!warning skipping  vertext cut"

        # learn # of pcs and generate on nonchild view of data
        if (
            max(k_values) >= nofam_snpreader.iid_count
            or max(k_values) >= nofam_snpreader.sid_count
        ):
            raise Exception(
                "The number of PCs search should be less than the # of rows and also the # of cols in the matrix after near relatives are removed"
            )
        randomstate = 1

        from sklearn.model_selection import KFold

        n_folds = 10
        folds = KFold(n_splits=n_folds, shuffle=True, random_state=randomstate).split(
            list(range(nofam_snpreader.sid_count))
        )

        scores = np.zeros((k_values.shape[0], n_folds))
        for i_fold, [train_idx, test_idx] in enumerate(folds):
            Utr, Str, Vtr, mean = None, None, None, None
            logging.info("test set size: {0}".format(len(test_idx)))

            logging.info("creating X_train for fold {0}".format(i_fold))
            t0 = time.time()
            X_train = (
                nofam_snpreader[:, train_idx].read(order="F").val.T
            )  #!!would order='C' be faster?
            logging.info("done after %.4f seconds" % (time.time() - t0))

            logging.info("creating X_test for fold {0}".format(i_fold))
            t0 = time.time()
            X_test = nofam_snpreader[:, test_idx].read(order="F").val.T
            logging.info("done after %.4f seconds" % (time.time() - t0))

            logging.info("Creating svd")
            t0 = time.time()
            for i_k, k in enumerate(k_values):
                pca = PCA(n_components=k, copy=False)
                Utr, Str, Vtr, mean = pca._fit(X_train, Utr, Str, Vtr, mean)
                if t0 is not None:
                    logging.info("done after %.4f seconds" % (time.time() - t0))
                    t0 = None
                scores[i_k, i_fold] = pca.score(X_test)
                logging.info("{0},{1},{2}".format(i_fold, k, scores[i_k, i_fold]))

        normalizedMean = scores.mean(axis=1)

        logging.info("normalized Means: {0}".format(normalizedMean))

        bestNumPCs = k_values[normalizedMean.argmax()]
        logging.info("best num PCs: {0}".format(bestNumPCs))

        logging.info("computing svd...")
        t0 = time.time()
        Utr, Str, Vtr, mean = None, None, None, None
        pca = PCA(n_components=bestNumPCs, copy=False)
        nofam_snpdata = nofam_snpreader.read(order="F")  #!!would order='C' be faster?
        pca._fit(
            nofam_snpdata.val, Utr, Str, Vtr, mean
        )  # This will zero-center nofam_snpdata.val, so don't use after this point.
        t1 = time.time()
        logging.info("done after %.4f seconds" % (t1 - t0))

        # apply those pcs to all the data (i.e.  transform)
        logging.info("Projecting individuals to PCs space...")
        X_fit = pca.transform(all_std_snpdata.val)

        ##write the file out
        if output_file_name is not None:
            logging.info("writing results to file...")
            with open(output_file_name, "w") as f:
                for iid_index, (famid, indid) in enumerate(all_std_snpdata.iid):
                    f.write("{0} {1} ".format(famid, indid))
                    f.write(" ".join([str(pc) for pc in X_fit[iid_index, :]]))
                    f.write("\n")

        result = {
            "iid": np.array(snpreader.iid),
            "vals": X_fit,
            "header": ["pc_{0}".format(index) for index in range(bestNumPCs)],
        }
        return result


def _snp_fixup(snp_input, count_A1=None):
    if isinstance(snp_input, str):
        return Bed(snp_input, count_A1=count_A1)
    else:
        return snp_input


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    print("done")



import numpy as np
import subprocess, sys, os.path
import fastlmm.util.util as utilx


class RandomSnpSet(object):  # implements ISnpSet
    """
    See the Bed class's 'read' method of examples of its use.
    See __init__.py for specification of interface it implements.
    """

    def __init__(self, numsnps, randomseed):
        self.numsnps = numsnps
        self.randomseed = randomseed

    def addbed(self, bed):
        return RandomSnpSetPlusBed(self.numsnps, self.randomseed, bed)


class RandomSnpSetPlusBed(object):  # implements ISnpSetPlusBed
    """
    One random set of snps.
    """

    def __init__(self, numsnps, randomseed, bed):
        self.numsnps = numsnps
        self.randomseed = randomseed
        self.bed = bed
        # very inefficient extraction of a few random SNP indexes:
        self.snpindlist = utilx.generate_permutation(
            np.arange(0, bed.snp_count), randomseed
        )[0:numsnps]
        self.snpindlist.sort()

    def __str__(self):
        return "RandomSnpSet(numsnps={0},randomseed={1})".format(
            self.numsnps, self.randomseed
        )

    def __len__(self):
        return self.numsnps

    def __iter__(self):
        for i in self.snpindlist:
            yield i

    def read(self):  #!!why don't all the interface implementers have this method?
        return self.bed.read_with_specification(self)

    @property
    def pos(self):
        """
        Returns:
            pos:    position of the SNPs in the specification
        """
        return self.bed.pos[self.to_index]

    @property
    def to_index(self):
        iter = self.__iter__()
        return [i for i in iter]
